import copy

import numpy as np
import stim

import qec
from qec import LogicalCode, tools
from typing import List, Union, Optional, Callable, Iterable, Sized
import random


class LogicalCircuit(stim.Circuit):
    def __init__(self, logical_qubits: List[LogicalCode],
                gate_noise: Optional[
                    Callable[[Union[stim.CircuitInstruction, stim.CircuitRepeatBlock]], stim.Circuit]] = None,
                idle_noise: Optional[
                    Callable[[Union[stim.CircuitInstruction, stim.CircuitRepeatBlock]], stim.Circuit]] = None,
                idle_noise_scale_factor: float = 1.0,
                gate_noise_scale_factor: float = 1.0,
                loss_noise_scale_factor: float = 1.0,
                spam_noise_scale_factor: float = 1.0,
                idle_loss_rate: float = 1e-7,
                idle_error_rate: tuple = (5e-6/25, 5e-6/25, 2e-5/25),
                entangling_zone_error_rate: tuple = (.002/4, .002/4, .005/4),
                entangling_gate_error_rate: tuple = (.002 / 4, .002 / 4, .0025 / 4, .002 / 4, 0, 0, 0, .002 / 4, 0, 0, 0, .0025 / 4, 0, 0, .005 / 4), 
                erasure_ratio: float = 0.0,
                entangling_gate_loss_rate: float = .005/4,
                single_qubit_loss_rate: float = 0.0,
                single_qubit_error_rate: tuple = (1e-4, 1e-4, 1e-4),
                reset_error_rate: float = 0.003, measurement_error_rate: float = 0.004,
                reset_loss_rate: float = 0,
                initialize_circuit: bool = True,
                atom_array_sim: bool = False):
        """
        Generates a LogicalCircuit, which is a subclass of stim.Circuit() but with additional attributes that are
        atom-array specific.
        :param logical_qubits: A list of qec.LogicalCode objects corresponding to the logical qubits in the circuit.
        :param gate_noise: A function that takes in a stim CircuitInstruction, and decorates it with noise (defaults
        to self.standard_gate_noise)
        :param idle_noise: A function that takes in a stim CircuitInstruction, and decorates it with noise (defaults
        to self.standard_idle_noise)
        :param idle_loss_rate: The rate at which qubits are lost during atom movements, in loss per qubit per us)
        :param idle_error_rate: The rate at which qubits accrue errors during movements, in error probability per qubit
        per us)
        :param entangling_gate_error_rate: The rate at which qubits accrue errors during entangling operations, in error
        probability per qubit per us)
        :param entangling_gate_loss_rate: The rate at which qubits are lost during an entangling operation
        :param single_qubit_loss_rate: The rate at which qubits are lost during a single-qubit gate operation
        :param single_qubit_error_rate:
        :param reset_error_rate: The probability that a qubit is initialized in |1> instead of |0>
        :param measurement_error_rate: The probability that an atom is flipped before it is measured in the Z basis
        :param reset_loss_rate: The probability that an atom is lost when it is reset (initialized in |0>)
        :param initialize_circuit: True if qubit coordinate initialization is done at the beginning of the circuit
        """
        super().__init__()
        self.logical_qubits = list(logical_qubits)
        self.num_logical_qubits = len(self.logical_qubits)
        self.logical_qubit_indices = np.arange(self.num_logical_qubits)
        LogicalCircuit.gate_noise = gate_noise
        LogicalCircuit.idle_noise = idle_noise
        
        self.atom_array_sims = atom_array_sim # if True: we work with zones. False: no zones, only errors after operations on the targets
        
        if gate_noise is None:
            if self.atom_array_sims:
                self.gate_noise = self.standard_gate_noise
            else:
                self.gate_noise = self.biased_erasure_entangling_noise
        if idle_noise is None:
            if self.atom_array_sims:
                self.idle_noise = self.standard_idle_noise
            else:
                self.idle_noise = self.no_idle_noise

        # Rename physical indices to account for multiple logical qubits
        for (_, logical_qubit) in enumerate(self.logical_qubits):
            if _ == 0:
                logical_qubit.qubit_indices = np.arange(logical_qubit.qubit_number)
            else:
                logical_qubit.qubit_indices = np.arange(self.logical_qubits[_ - 1].qubit_indices[-1] + 1,
                                                        self.logical_qubits[_ - 1].qubit_indices[-1] +
                                                        logical_qubit.qubit_number + 1)

        if len(self.logical_qubits) > 0:
            self.qubit_indices = np.concatenate([logical_qubit.qubit_indices
                                                for logical_qubit in self.logical_qubits])
        else:
            self.qubit_indices = np.array([])

        # Zones - relevant for atom array sims:
        self.entangling_zone = set()
        self.storage_zone = set()
        self.lost_qubits = set()
        self.no_noise_zone = set(self.qubit_indices)
        
        self.idle_noise_scale_factor = idle_noise_scale_factor
        self.gate_noise_scale_factor = gate_noise_scale_factor
        self.loss_noise_scale_factor = loss_noise_scale_factor
        self.spam_noise_scale_factor = spam_noise_scale_factor
        self.idle_loss_rate = idle_loss_rate
        self.idle_error_rate = idle_error_rate
        self.entangling_zone_error_rate = entangling_zone_error_rate
        self.entangling_gate_error_rate = entangling_gate_error_rate
        self.entangling_gate_loss_rate = entangling_gate_loss_rate
        self.erasure_ratio = erasure_ratio
        self.reset_error_rate = reset_error_rate
        self.reset_loss_rate = reset_loss_rate
        self.single_qubit_error_rate = single_qubit_error_rate
        self.single_qubit_loss_rate = single_qubit_loss_rate
        self.measurement_error_rate = measurement_error_rate
        self._without_loss = stim.Circuit()

        # Handling loss:
        self.potential_lost_qubits = np.array([], dtype=np.int32)
        self.loss_probabilities = np.array([], dtype=np.float32)
        
        
        
        if initialize_circuit:
            self.initialize_circuit()

    def initialize_circuit(self, circuit: Optional[stim.Circuit] = None):
        if circuit is None:
            for logical_qubit in self.logical_qubits:
                super(LogicalCircuit, self).__iadd__(logical_qubit.qubit_coordinates())
        else:
            for logical_qubit in self.logical_qubits:
                circuit += logical_qubit.qubit_coordinates()
            return circuit

    def qubit_index_to_logical_qubit_index(self, index: int):
        for (_, lq) in enumerate(self.logical_qubits):
            if index in lq.qubit_indices:
                return _
        raise Exception('No logical qubit found with corresponding physical index')

    def biased_erasure_entangling_noise(self, operation: stim.CircuitInstruction, add_to_potential_loss: bool = True) -> stim.Circuit:
        """
        GB: biased_erasure noise channel - noise after 2q gates.
        """
        noisy_circuit = stim.Circuit()
        noisy_circuit.append(operation)
        if qec.is_entangling(operation.name):
            targets = [t.value for t in operation.targets_copy()]
            
            # put errors only on data qubits - debugging:
            # data_qubits = []
            # for lq in self.logical_qubits:
            #     data_qubits.extend(lq.data_qubits)
            # targets = [q for q in targets if q in data_qubits]
            
            # put errors only on ancilla qubits - debugging:
            # ancilla_qubits = []
            # for lq in self.logical_qubits:
            #     ancilla_qubits.extend(lq.measure_qubits)
            # targets = [q for q in targets if q in ancilla_qubits]
            
            
            # 2nd approach - put I channel all the time, to not make mistakes with the SWAP
            # Randomly sample which qubit gets loss and which qubit gets Pauli.
            err_types = random.choices(['loss','pauli'], weights=[self.erasure_ratio,(1-self.erasure_ratio)], k=int(len(targets)))

            # Pauli noise channel:
            pauli_targets = [target for target, err_type in zip(targets, err_types) if err_type == 'pauli'] # take all targets with 'pauli' errors:
            pauli_error = np.array(self.entangling_gate_error_rate) * self.gate_noise_scale_factor
            if sum(pauli_error) > 0 and len(pauli_targets)> 0:
                if len(pauli_error) == 3:
                    noisy_circuit.append('PAULI_CHANNEL_1', pauli_targets, pauli_error)
                elif len(pauli_error) == 8:
                    noisy_circuit.append('PAULI_CHANNEL_2', pauli_targets, pauli_error)
            
            # 1q loss channel on each qubit that did the gate: (TODO: change this to put I channel on all qubits "targets" but adjust the self.loss_probabilities to be loss_prob multiply by 1 if 'loss' on this qubit or 0 is 'pauli' on this qubit
            loss_targets = [target for target, err_type in zip(targets, err_types) if err_type == 'loss'] # take all other targets:
            loss_prob = self.entangling_gate_loss_rate * self.loss_noise_scale_factor
            if add_to_potential_loss and loss_prob>0:
                # Apply I channel to all targets to simulate identity operation
                noisy_circuit.append('I', targets)

                # Update the potential lost qubits list and loss probabilities
                for target, err_type in zip(targets, err_types):
                    if err_type == 'loss':
                        loss_prob = self.entangling_gate_loss_rate * self.loss_noise_scale_factor
                        self.potential_lost_qubits = np.append(self.potential_lost_qubits, target)
                        self.loss_probabilities = np.append(self.loss_probabilities, loss_prob)
                    else:
                        # If it's not a loss, append zero to maintain the correct indices
                        self.potential_lost_qubits = np.append(self.potential_lost_qubits, target)
                        self.loss_probabilities = np.append(self.loss_probabilities, 0)

            
        return noisy_circuit

    def standard_gate_noise(self, operation: stim.CircuitInstruction, add_to_potential_loss:bool = False) -> stim.Circuit:
        assert self.atom_array_sims
        
        noisy_circuit = stim.Circuit()

        # Add measurement noise before operation
        if qec.is_measurement(operation.name):
            measurement_noise = [t.value for t in operation.targets_copy() if t.value not in self.no_noise_zone]
            if len(measurement_noise) != 0:
                noisy_circuit.append('X_ERROR', measurement_noise,
                                     [self.measurement_error_rate * self.spam_noise_scale_factor])

        noisy_circuit.append(operation)

        # Add other errors after operation
        if qec.is_entangling(operation.name):
            # Add noise to all qubits in the entangling zone
            targets = [t.value for t in operation.targets_copy() if t.value not in self.no_noise_zone]
            # qubits that undergo an entangling gate, get a 2q channel:
            noisy_circuit.append('PAULI_CHANNEL_2', targets,
                                np.array(self.entangling_gate_error_rate) * self.gate_noise_scale_factor)
            # qubits that don't undergo an entangling gate but are in the entangling zone, get a 1q channel:
            residual_entangling_zone_qubits = self.entangling_zone - set(targets)
            noisy_circuit.append('PAULI_CHANNEL_1', residual_entangling_zone_qubits,
                                np.array(self.entangling_zone_error_rate) * self.gate_noise_scale_factor)
            # qubits in the entangling zone also get loss:
            qubits = np.setdiff1d(list(self.entangling_zone), list(self.no_noise_zone)) # entangling zone minus no noise zone
            loss_prob = self.entangling_gate_loss_rate * self.loss_noise_scale_factor
            if loss_prob > 0 and add_to_potential_loss:
                noisy_circuit.append('I', qubits)
                self.potential_lost_qubits = np.append(self.potential_lost_qubits, qubits)
                self.loss_probabilities = np.append(self.loss_probabilities, np.full(len(qubits), loss_prob))     


        if qec.is_reset(operation.name):
            # noisy_circuit.append('X_ERROR', list(self.entangling_zone), [self.reset_error_rate * self.spam_noise_scale_factor])
            noisy_circuit.append('X_ERROR', list(operation.targets_copy()), [self.reset_error_rate * self.spam_noise_scale_factor]) # GB's new change - only initialized qubits get the error
            qubits = np.setdiff1d(list(self.qubit_indices), list(self.no_noise_zone))
            loss_prob = self.reset_loss_rate * self.loss_noise_scale_factor
            if loss_prob > 0 and add_to_potential_loss:
                noisy_circuit.append('I', qubits)
                self.potential_lost_qubits = np.append(self.potential_lost_qubits, qubits)
                self.loss_probabilities = np.append(self.loss_probabilities, np.full(len(qubits), loss_prob))


        if qec.is_single_qubit(operation.name):
            # Add single qubit noise only on the qubits we do the operation on
            noisy_circuit.append('PAULI_CHANNEL_1', [t.value for t in operation.targets_copy() if t.value
                                                    not in self.no_noise_zone],
                                 np.array(self.single_qubit_error_rate) * self.gate_noise_scale_factor)
            qubits = np.setdiff1d(list(self.qubit_indices), list(self.no_noise_zone))
            loss_prob = self.single_qubit_loss_rate * self.loss_noise_scale_factor
            if loss_prob > 0 and add_to_potential_loss:
                noisy_circuit.append('I', qubits) # loss probability: self.entangling_gate_loss_rate * self.loss_noise_scale_factor
                self.potential_lost_qubits = np.append(self.potential_lost_qubits, qubits)
                self.loss_probabilities = np.append(self.loss_probabilities, np.full(len(qubits), loss_prob))
            
        return noisy_circuit
    
    
    def standard_idle_noise(self, duration: Union[float, int], add_to_potential_loss:bool) -> stim.Circuit:
        noisy_circuit = stim.Circuit()
        if duration != 0:
            # noisy_circuit.append('TICK', ())
            noisy_circuit.append('PAULI_CHANNEL_1', self.entangling_zone | self.storage_zone, (1 -
                                 (1 - np.array(self.idle_error_rate)) ** duration) * self.idle_noise_scale_factor)
            # Add per qubit loss noise
            loss_prob = (1 - (1 - self.idle_loss_rate) ** duration) * self.loss_noise_scale_factor
            qubits = np.setdiff1d(list(self.qubit_indices), list(self.no_noise_zone))
            if loss_prob > 0 and add_to_potential_loss:
                noisy_circuit.append('I', qubits) # loss probability: self.entangling_gate_loss_rate * self.loss_noise_scale_factor
                self.potential_lost_qubits = np.append(self.potential_lost_qubits, qubits)
                self.loss_probabilities = np.append(self.loss_probabilities, np.full(len(qubits), loss_prob))

        return noisy_circuit


    def add_idle_noise(self, duration: Union[float, int], add_to_potential_loss: bool) -> stim.Circuit:
        return self.idle_noise(duration, add_to_potential_loss)
    
    def no_idle_noise(self, duration: Union[float, int]) -> stim.Circuit:
        return stim.Circuit()
    

    def add_noise(self, circuit: stim.Circuit, add_to_potential_loss: bool) -> stim.Circuit:
        noisy_circuit = stim.Circuit()
        for operation in circuit:
            if isinstance(operation, stim.CircuitRepeatBlock):
                noisy_circuit += self.add_noise(operation.body_copy(), add_to_potential_loss)
            else:
                assert isinstance(operation, stim.CircuitInstruction)
                noisy_circuit += self.gate_noise(operation, add_to_potential_loss)
        return noisy_circuit

    def _generate_detector(self, pauli: stim.PauliString, noiseless_circuit=None):
        # Look at the last time that the measure qubit was measured
        if noiseless_circuit is None:
            noiseless_circuit = self.without_noise()
        #print('New')
        #print(pauli)
        n_measurements = len(noiseless_circuit[-1].targets_copy())
        c_t = stim.Circuit()
        c_t = self.initialize_circuit(c_t)
        # Generate Pauli measurement corresponding to current stabilizer
        recs = []
        for i in range(2, len(noiseless_circuit)):
            if qec.is_measurement(noiseless_circuit[-i].name):
                # Find the measured stabilizers
                # measured_qubits = [target.value for target in noiseless_circuit[-i].targets_copy()]
                if noiseless_circuit[-i].name == 'MPP':
                    measured_qubits = np.array([target.value for target in noiseless_circuit[-i].targets_copy()])[::2]
                else:
                    measured_qubits = [target.value for target in noiseless_circuit[-i].targets_copy()]
                    
                # Find the propagated stabilizer
                t = stim.Tableau.from_circuit(c_t, ignore_measurement=True, ignore_noise=True, ignore_reset=True)
                propagated_stabilizer = np.array(t(stim.PauliString(pauli)))
                new_pauli = propagated_stabilizer.copy()
                #print(stim.PauliString(new_pauli))
                measured_stabilizers = []
                propagated_stabilizers = []
                # Figure out what stabilizers we propagated to
                for lq in self.logical_qubits:
                    for mq in lq.measure_qubits:
                        dq = [d for d in lq.measure_to_data[mq] if d is not None]
                        if mq in lq.measure_qubits_x:
                            if np.alltrue(np.isin(propagated_stabilizer[dq], [1, 2])):
                                propagated_stabilizers.append(mq)
                                if mq in measured_qubits:
                                    measured_stabilizers.append(mq)
                                    for d in dq:
                                        if new_pauli[d] == 1:
                                            new_pauli[d] = 0
                                        elif new_pauli[d] == 2:
                                            new_pauli[d] = 3

                        elif mq in lq.measure_qubits_z:
                            if np.alltrue(np.isin(propagated_stabilizer[dq], [3, 2])):
                                propagated_stabilizers.append(mq)
                                if mq in measured_qubits:
                                    measured_stabilizers.append(mq)
                                    for d in dq:
                                        if new_pauli[d] == 3:
                                            new_pauli[d] = 0
                                        elif new_pauli[d] == 2:
                                            new_pauli[d] = 1
                        new_pauli[mq] = 0

                # Now we need to find the corresponding measurement record
                # Because we've measured these stabilizers, we want to multiply them into the tableau to cancel
                # them out.
                # n_measurements += len(noiseless_circuit[-i].targets_copy())
                if noiseless_circuit[-i].name == 'MPP':
                    n_measurements += 1
                else:
                    n_measurements += len(noiseless_circuit[-i].targets_copy())
                #print(stim.PauliString(new_pauli))
                if len(measured_stabilizers) > 0:
                    for (_, s) in enumerate(measured_stabilizers):
                        if s in measured_qubits:
                            recs.append(-(n_measurements +
                                        (self.without_noise().num_measurements - noiseless_circuit.num_measurements)
                                        - np.argwhere(measured_qubits == s)[0, 0]))
                    if np.all(new_pauli == np.zeros_like(propagated_stabilizer)):
                        return recs
                    #elif np.all(new_pauli == pauli) and np.sum(pauli!=0):
                    #    break
                    else:
                        recs = recs + self._generate_detector(stim.PauliString(new_pauli),
                                                            noiseless_circuit=noiseless_circuit[:-i + 1])
                        break
            c_t_current = stim.Circuit()
            c_t_current.append(noiseless_circuit[-i])
            # Try to add the inverse instruction. If it's not possible, then it's a measurement, reset, or detector.
            # Then we just want to add the circuit element
            try:
                c_t += (c_t_current.inverse())
            except:
                c_t += c_t_current

        return recs

    def append(self, name: object, targets: object = (), arg: object = None, **kwargs) -> None:
        if callable(name):
            name(self, targets, **kwargs)

        # Handle noise from movement
        elif name in ['MOVE', 'MOVE_TO_STORAGE', 'MOVE_TO_ENTANGLING', 'MOVE_TO_NO_NOISE']:
            if not self.atom_array_sims:
                pass
            else:
                assert isinstance(targets, int) or isinstance(targets, Iterable)
                assert isinstance(arg, int) or isinstance(arg, float) or arg is None
                if not(arg is None):
                    if name == 'MOVE_TO_STORAGE':
                        self.entangling_zone = self.entangling_zone - set(targets)
                        self.no_noise_zone = self.no_noise_zone - set(targets)
                        self.storage_zone = self.storage_zone | set(targets)

                        # Check if adding noise from the movement is actually necessary 
                        if not set(targets).issubset(self.storage_zone): #GB: isn't this always False? because we just more the targets into the storage_zone
                            super().__iadd__(self.add_idle_noise(arg, add_to_potential_loss = True))
                            idle_loss_rate, self.idle_loss_rate = self.idle_loss_rate, 0.
                            self._without_loss += self.add_idle_noise(arg, add_to_potential_loss = False)
                            self.idle_loss_rate = idle_loss_rate

                    elif name == 'MOVE_TO_ENTANGLING':
                        self.entangling_zone = self.entangling_zone | set(list(targets))
                        self.no_noise_zone = self.no_noise_zone - set(list(targets))
                        self.storage_zone = self.storage_zone - set(list(targets))

                        # Check if adding noise from the movement is actually necessary 
                        if not set(list(targets)).issubset(self.entangling_zone): #GB: isn't this always False? because we just more the targets into the entangling_zone
                            super().__iadd__(self.add_idle_noise(arg, add_to_potential_loss = True))
                            idle_loss_rate, self.idle_loss_rate = self.idle_loss_rate, 0.
                            self._without_loss += self.add_idle_noise(arg, add_to_potential_loss = False)
                            self.idle_loss_rate = idle_loss_rate

                    elif name == 'MOVE_TO_NO_NOISE':
                        self.no_noise_zone = self.no_noise_zone | set(list(targets))
                        self.entangling_zone = self.entangling_zone - set(list(targets))
                        self.storage_zone = self.storage_zone - set(list(targets))

                        # Check if adding noise from the movement is actually necessary
                        if not set(list(targets)).issubset(self.entangling_zone): #GB: why do we check this here?
                            super().__iadd__(self.add_idle_noise(arg, add_to_potential_loss = True))
                            idle_loss_rate, self.idle_loss_rate = self.idle_loss_rate, 0.
                            self._without_loss += self.add_idle_noise(arg, add_to_potential_loss = False)
                            self.idle_loss_rate = idle_loss_rate

                    else:
                        super().__iadd__(self.add_idle_noise(arg, add_to_potential_loss = True))
                        idle_loss_rate, self.idle_loss_rate = self.idle_loss_rate, 0.
                        self._without_loss += self.add_idle_noise(arg, add_to_potential_loss = False)
                        self.idle_loss_rate = idle_loss_rate
            
        # Add detectors for stabilizer measurements
        elif name in ['MEASURE_STABILIZERS']:
            # Assert that the previous circuit instruction was to measure the stabilizers
            if not qec.is_measurement((self.without_noise())[-1].name):
                raise Exception('MEASURE_STABILIZERS command must follow a measurement of stabilizers.')

            measured_stabilizers = targets
            measured_qubits = [target.value for target in (self.without_noise())[-1].targets_copy()]
            assert self.without_noise()[-1].name in ['M', 'MZ', 'MRZ', 'MR']

            # Default to False if compare_with_previous not specified
            if not ('compare_with_previous' in kwargs):
                kwargs['compare_with_previous'] = False
            circuit = stim.Circuit()
            if not kwargs['compare_with_previous']:
                circuit.append('DETECTOR', [stim.target_rec(-i) for i in range(1, len(measured_qubits) + 1)])
            else:
                if isinstance(measured_stabilizers, int):
                    measured_stabilizers = [measured_stabilizers]
                assert isinstance(measured_stabilizers, Iterable)
                # We need to go back through our circuit and match with the previous stabilizer measurement
                # First determine if we've measured stabilizers or done a round of perfect error correction
                all_recs = []
                for stabilizer in measured_stabilizers:
                    # Find the logical qubit and corresponding data qubits for this measure qubit
                    logical_qubit = self.qubit_index_to_logical_qubit(stabilizer)
                    data_qubits = [d for d in logical_qubit.measure_to_data[stabilizer] if d is not None]

                    if stabilizer in measured_qubits and np.all([d in measured_qubits for d in data_qubits]):
                        # If we did a round of perfect error correction and measured stabilizers simultaneously, we
                        # pick this branch
                        recs = []
                        # Add the measure qubit measurement record
                        matches = np.argwhere(measured_qubits == stabilizer)
                        if len(matches) > 0:
                            recs.append(-(len(measured_qubits) - matches[0, 0]))
                        # Find the correspond data qubit measurement records
                        for data_qubit in data_qubits:
                            matches = np.argwhere(measured_qubits == data_qubit)
                            recs.append(-(len(measured_qubits) - matches[0, 0]))

                    elif stabilizer in measured_qubits:
                        # If we just measured a measure qubit, we pick this branch
                        recs = []
                        matches = np.argwhere(measured_qubits == stabilizer)
                        if len(matches) > 0:
                            recs.append(-(len(measured_qubits)-matches[0,0]))
                        # If the ancilla qubit was measured we pick this branch
                        pauli = np.zeros(self.num_qubits, dtype=int)
                        # Assume we've measured in the Z-basis
                        pauli[stabilizer] = 3
                        recs = recs + self._generate_detector(stim.PauliString(pauli))

                    else:
                        # If only the data qubits were measured we pick this branch
                        # Search for these qubits in
                        if not all([d in measured_qubits for d in data_qubits]):
                            raise Exception('The data qubits corresponding to the stabilizers were not measured.')

                        recs = []
                        for (m, qubit) in enumerate(measured_qubits):
                            if qubit in data_qubits:
                                recs.append(-(len(measured_qubits) - m))
                        pauli = np.zeros(self.num_qubits, dtype=int)
                        logical_qubit = self.qubit_index_to_logical_qubit(stabilizer)
                        data_qubit_indices = [d for d in logical_qubit.measure_to_data[stabilizer] if d is not None]
                        measurement_operation = self.without_noise()[-1].name
                        #print(stabilizer, data_qubit_indices)
                        if measurement_operation in ['M', 'MR', 'MZ', 'MRZ']:
                            # Measure in the Z basis
                            pauli[data_qubit_indices] = 3
                        elif measurement_operation in ['MX', 'MRX']:
                            # Measure in the X basis
                            pauli[data_qubit_indices] = 1
                        elif measurement_operation in ['MY', 'MRY']:
                            # Measure in the X basis
                            pauli[data_qubit_indices] = 2
                        else:
                            # We measured a Pauli product. We need to deal with this. But another day.
                            raise NotImplementedError('Automatic detectors not implemented for MPP measurements.')
                        recs = recs + self._generate_detector(stim.PauliString(pauli))
                    all_recs.append([stim.target_rec(r) for r in recs])

                for recs in all_recs:
                    self.append('DETECTOR', recs)

            super().__iadd__(circuit)
            self._without_loss += circuit

        # Handle other circuit noise
        else:
            assert isinstance(targets, Sized)

            circuit = stim.Circuit()
            circuit.append(name, targets, arg)
            entangling_gate_loss_rate, reset_loss_rate, single_qubit_loss_rate = self.entangling_gate_loss_rate, self.reset_loss_rate, self.single_qubit_loss_rate
            self.entangling_gate_loss_rate, self.reset_loss_rate, self.single_qubit_loss_rate = 0., 0., 0.
            self._without_loss += self.add_noise(circuit, add_to_potential_loss = False)
            self.entangling_gate_loss_rate, self.reset_loss_rate, self.single_qubit_loss_rate = entangling_gate_loss_rate, reset_loss_rate, single_qubit_loss_rate

            # Add only circuit instructions on non-lost qubits # NOT RELEVANT ANYMORE, if we want to make it faster we can just append the instruction to all targets
            circuit = stim.Circuit()
            if tools.is_entangling(str(name)):
                # Remove both qubits in the gate if a single qubit in the entangling gate is lost
                not_lost_targets = []
                for _ in range(len(targets) // 2):
                    # We always want to do swaps, because it corresponds to when we physically rotate the code!
                    if (not (targets[2 * _] in self.lost_qubits) and not (targets[2 * _ + 1] in self.lost_qubits)) \
                            or str(name) == 'SWAP':
                        not_lost_targets.append(targets[2 * _])
                        not_lost_targets.append(targets[2 * _ + 1])
                circuit.append(name, not_lost_targets, arg)
                super().__iadd__(self.add_noise(circuit, add_to_potential_loss=True))
            elif tools.is_single_qubit(str(name)):
                circuit.append(name, targets, arg)
                super().__iadd__(self.add_noise(circuit, add_to_potential_loss=True))
            elif tools.is_measurement(str(name)):
                # Reset lost qubits in |1>
                reset_circuit = stim.Circuit()
                lost_targets = [t for t in targets if t in self.lost_qubits]
                if len(lost_targets) > 0:
                    reset_circuit.append('R', lost_targets)
                    reset_circuit.append('X_ERROR', lost_targets, 1)
                    super().__iadd__(reset_circuit, add_to_potential_loss=True)

                self.lost_qubits = self.lost_qubits - set(lost_targets)
                circuit.append(name, targets, arg)
                super().__iadd__(self.add_noise(circuit, add_to_potential_loss=True))
            
            elif tools.is_reset(str(name)): # GB's addition
                circuit.append(name, targets, arg)
                super().__iadd__(self.add_noise(circuit, add_to_potential_loss=True))
            
            else:
                circuit.append(name, targets, arg)
                super().__iadd__(self.add_noise(circuit, add_to_potential_loss=True))

    def qubit_index_to_logical_qubit(self, index: int):
        for lq in self.logical_qubits:
            if index in lq.qubit_indices:
                return lq
        raise Exception('No logical qubit found with corresponding physical index')

    def without_loss(self) -> stim.Circuit:
        # print("This feature is not implemented anymore.")
        # return None
        return self._without_loss

    def without_noise(self) -> stim.Circuit:
        # print("This feature is not implemented anymore.")
        # return None
        return self.without_loss().without_noise()

    def copy(self, memodict={}):
        return copy.deepcopy(self, memo=memodict)



import numpy as np
import random
import os
import qec
import pymatching, stim
import random
import math
import time
import sinter
import copy
from BiasedErasure.main_code.XZZX import XZZX
from BiasedErasure.main_code.noise_channels import biased_erasure_noise_MBQC
from BiasedErasure.main_code.LogicalCircuitMBQC import LogicalCircuitMBQC
from BiasedErasure.main_code.LogicalCircuit import LogicalCircuit 

def memory_experiment_surface_new(d, code, QEC_cycles, entangling_gate_error_rate, entangling_gate_loss_rate, erasure_ratio, num_logicals=1, logical_basis='X', biased_pres_gates = False, ordering = 'fowler', loss_detection_method = 'FREE', loss_detection_frequency = 1, atom_array_sim=False):
    """ This circuit simulated 1 logical qubits, a memory experiment with QEC cycles. We take perfect initialization and measurement and put noise only on the QEC cycles part."""
    assert logical_basis in ['X', 'Z'] # init and measurement basis for the single qubit logical state
    print(f"entangling Pauli error rate = {entangling_gate_error_rate}, entangling loss rate = {entangling_gate_loss_rate}")
    
    if code == 'Rotated_Surface':
        logical_qubits = [qec.surface_code.RotatedSurfaceCode(d, d) for _ in range(num_logicals)]
    elif code == 'Surface':
        logical_qubits = [qec.surface_code.SurfaceCode(d, d) for _ in range(num_logicals)]
    
    if atom_array_sim:
        lc = LogicalCircuit(logical_qubits, initialize_circuit=False,
                                loss_noise_scale_factor=1, spam_noise_scale_factor=1,
                                gate_noise_scale_factor=1, idle_noise_scale_factor=1,
                                atom_array_sim = atom_array_sim)
    else:
        lc = LogicalCircuit(logical_qubits, initialize_circuit=False,
                                loss_noise_scale_factor=1, spam_noise_scale_factor=0,
                                gate_noise_scale_factor=1, idle_noise_scale_factor=0,
                                entangling_gate_error_rate=entangling_gate_error_rate, 
                                entangling_gate_loss_rate=entangling_gate_loss_rate,
                                erasure_ratio = erasure_ratio,
                                atom_array_sim = atom_array_sim)
    
    #  initialization step:
    if not atom_array_sim: lc.loss_noise_scale_factor = 0; lc.gate_noise_scale_factor=0
    
    if logical_basis == 'X':
        lc.append(qec.surface_code.prepare_plus_no_gates, list(range(0, len(logical_qubits))))
    
    elif logical_basis == 'Z':
        lc.append(qec.surface_code.prepare_zero_no_gates, list(range(0, len(logical_qubits))))
        
    if not atom_array_sim: lc.loss_noise_scale_factor = 1; lc.gate_noise_scale_factor=1
        
        
    # QEC rounds:
    SWAP_round_index = 0; SWAP_round_type = 'even'; SWAP_round = False
    for round_ix in range(QEC_cycles):
        if loss_detection_method == 'SWAP' and ((round_ix+1)%loss_detection_frequency == 0): # check if its a SWAP round:
            SWAP_round = True
            SWAP_round_type = 'even' if SWAP_round_index%2 ==0 else 'odd'
            SWAP_round_index += 1
        
        put_detectors = False if round_ix == 0 else True
        lc.append_from_stim_program_text("""TICK""") # starting a QEC round
        lc.append(qec.surface_code.measure_stabilizers, list(range(len(logical_qubits))), order=ordering, with_cnot=biased_pres_gates, SWAP_round = SWAP_round, SWAP_round_type=SWAP_round_type, compare_with_previous=True, put_detectors = put_detectors, logical_basis=logical_basis) # append QEC rounds
        lc.append_from_stim_program_text("""TICK""") # starting a QEC round
    
    
    # logical measurement step:
    if not atom_array_sim: lc.loss_noise_scale_factor = 0; lc.gate_noise_scale_factor=0
    
    if logical_basis == 'X':
        lc.append(qec.surface_code.measure_x, list(range(len(logical_qubits))), observable_include=True)
    
    elif logical_basis == 'Z':
        lc.append(qec.surface_code.measure_z, list(range(len(logical_qubits))), observable_include=True)
        

    return lc


def memory_experiment_surface(d, code, QEC_cycles, entangling_gate_error_rate, entangling_gate_loss_rate, erasure_ratio, num_logicals=1, logical_basis='X', biased_pres_gates = False, ordering = 'fowler', loss_detection_method = 'FREE', loss_detection_frequency = 1, atom_array_sim=False):
    """ This circuit simulated 1 logical qubits, a memory experiment with QEC cycles. We take perfect initialization and measurement and put noise only on the QEC cycles part."""
    assert logical_basis in ['X', 'Z'] # init and measurement basis for the single qubit logical state
    print(f"entangling Pauli error rate = {entangling_gate_error_rate}, entangling loss rate = {entangling_gate_loss_rate}")
    
    if code == 'Rotated_Surface':
        logical_qubits = [qec.surface_code.RotatedSurfaceCode(d, d) for _ in range(num_logicals)]
    elif code == 'Surface':
        logical_qubits = [qec.surface_code.SurfaceCode(d, d) for _ in range(num_logicals)]
    
    if atom_array_sim:
        lc = LogicalCircuit(logical_qubits, initialize_circuit=False,
                                loss_noise_scale_factor=1, spam_noise_scale_factor=1,
                                gate_noise_scale_factor=1, idle_noise_scale_factor=1,
                                atom_array_sim = atom_array_sim)
    else:
        lc = LogicalCircuit(logical_qubits, initialize_circuit=False,
                                loss_noise_scale_factor=1, spam_noise_scale_factor=0,
                                gate_noise_scale_factor=1, idle_noise_scale_factor=0,
                                entangling_gate_error_rate=entangling_gate_error_rate, 
                                entangling_gate_loss_rate=entangling_gate_loss_rate,
                                erasure_ratio = erasure_ratio,
                                atom_array_sim = atom_array_sim)
    
    if not atom_array_sim: lc.loss_noise_scale_factor = 0; lc.gate_noise_scale_factor=0
    
    if logical_basis == 'X':
        lc.append(qec.surface_code.prepare_plus, list(range(0, len(logical_qubits))), order=ordering, with_cnot=biased_pres_gates)
        if not atom_array_sim: lc.loss_noise_scale_factor = 1; lc.gate_noise_scale_factor=1
        
        SWAP_round_index = 0; SWAP_round_type = 'none'; SWAP_round = False
        for round_ix in range(QEC_cycles):
            if loss_detection_method == 'SWAP' and ((round_ix+1)%loss_detection_frequency == 0): # check if its a SWAP round:
                SWAP_round = True
                SWAP_round_type = 'even' if SWAP_round_index%2 ==0 else 'odd'
                SWAP_round_index += 1
            
            lc.append_from_stim_program_text("""TICK""") # starting a QEC round
            lc.append(qec.surface_code.measure_stabilizers, list(range(len(logical_qubits))), order=ordering, with_cnot=biased_pres_gates, compare_with_previous=True, SWAP_round = SWAP_round, SWAP_round_type=SWAP_round_type) # append QEC rounds
            lc.append_from_stim_program_text("""TICK""") # starting a QEC round
        if not atom_array_sim: lc.loss_noise_scale_factor = 0; lc.gate_noise_scale_factor=0
        lc.append(qec.surface_code.measure_x, list(range(len(logical_qubits))), observable_include=True)

    elif logical_basis == 'Z':
        lc.append(qec.surface_code.prepare_zero, list(range(0, len(logical_qubits))), order=ordering, with_cnot=biased_pres_gates)
        if not atom_array_sim: lc.loss_noise_scale_factor = 1; lc.gate_noise_scale_factor=1
        
        SWAP_round_index = 0; SWAP_round_type = 'none'; SWAP_round = False
        for round_ix in range(QEC_cycles):
            if loss_detection_method == 'SWAP' and ((round_ix+1)%loss_detection_frequency == 0): # check if its a SWAP round:
                SWAP_round = True
                SWAP_round_type = 'even' if SWAP_round_index%2 ==0 else 'odd'
                SWAP_round_index += 1
        
            lc.append_from_stim_program_text("""TICK""") # starting a QEC round
            lc.append(qec.surface_code.measure_stabilizers, list(range(len(logical_qubits))), order=ordering, with_cnot=biased_pres_gates, compare_with_previous=True, SWAP_round = SWAP_round, SWAP_round_type=SWAP_round_type) # append QEC rounds
            lc.append_from_stim_program_text("""TICK""") # starting a QEC round
        if not atom_array_sim: lc.loss_noise_scale_factor = 0; lc.gate_noise_scale_factor=0
        lc.append(qec.surface_code.measure_z, list(range(len(logical_qubits))), observable_include=True)
    
    return lc

def memory_experiment_MBQC(d, QEC_cycles, entangling_gate_error_rate, entangling_gate_loss_rate, erasure_ratio, logical_basis='X', biased_pres_gates = False, atom_array_sim=False):
    """ This circuit simulated 1 logical qubits, a memory experiment with QEC cycles. We take perfect initialization and measurement and put noise only on the QEC cycles part.
    We are using the XZZX cluster state. cycles=c --> 2*c + 1 layers in the cluster state.
    
    """
    xzzx_instance = XZZX()  # Assuming XZZX has a constructor that can be called without arguments

    init_circuit = xzzx_instance.get_circuit(cycles=QEC_cycles, dx=d, dy=d, basis=logical_basis,
                                        offset=0, architecture = 'MBQC', bias_preserving_gates = biased_pres_gates, 
                                        atom_array_sim = atom_array_sim)
    
    # TODO: check bug
    if atom_array_sim:
        print("Need to build this feature in the new framework. It works only in the old framework.")
    else:
        noisy_circuit, loss_probabilities, potential_lost_qubits = biased_erasure_noise_MBQC(init_circuit, dx=d, dy=d,
                                        entangling_gate_error_rate=entangling_gate_error_rate, entangling_gate_loss_rate=entangling_gate_loss_rate, 
                                        erasure_weight=erasure_ratio, bias_preserving = biased_pres_gates)

    lc = LogicalCircuitMBQC(noisy_circuit, loss_probabilities, potential_lost_qubits)

    return lc




def Steane_QEC_circuit(d, code, Steane_type, QEC_cycles, entangling_gate_error_rate, entangling_gate_loss_rate, erasure_ratio, logical_basis='X',
                    biased_pres_gates = False, loss_detection_on_all_qubits = True, atom_array_sim=False, ancilla1_for_preselection: bool = False,
                    ancilla2_for_preselection: bool = False):
    assert logical_basis in ['X', 'Z'] # measurement basis for final logical state
    assert Steane_type in ['Regular', 'SWAP'] # measurement basis for final logical state
    num_logicals = 3
    # Steane_type can be 'Regular' or 'swap'.
    # q0: logical. q1: |+> ancilla. q2: |0> ancilla.
    
    if code == 'Rotated_Surface':
        logical_qubits = [qec.surface_code.RotatedSurfaceCode(d, d) for _ in range(num_logicals)]
    elif code == 'Surface':
        logical_qubits = [qec.surface_code.SurfaceCode(d, d) for _ in range(num_logicals)]
    
    if atom_array_sim:
        lc = LogicalCircuit(logical_qubits, initialize_circuit=False,
                                loss_noise_scale_factor=1, spam_noise_scale_factor=1,
                                gate_noise_scale_factor=1, idle_noise_scale_factor=1,
                                atom_array_sim = atom_array_sim)
    else:
        lc = LogicalCircuit(logical_qubits, initialize_circuit=False,
                                loss_noise_scale_factor=1, spam_noise_scale_factor=0,
                                gate_noise_scale_factor=1, idle_noise_scale_factor=0,
                                entangling_gate_error_rate=entangling_gate_error_rate, 
                                entangling_gate_loss_rate=entangling_gate_loss_rate,
                                erasure_ratio = erasure_ratio,
                                atom_array_sim = atom_array_sim)
        
    if not atom_array_sim: lc.loss_noise_scale_factor = 0; lc.gate_noise_scale_factor=0
    
    # First prepare all logical qubits:
    if logical_basis == 'X':
        lc.append(qec.surface_code.prepare_plus, [0])
    elif logical_basis == 'Z':
        lc.append(qec.surface_code.prepare_zero, [0])
    lc.append(qec.surface_code.prepare_plus, [1])
    lc.append(qec.surface_code.prepare_zero, [2])

    # FT preparation of logical ancilla qubits:
    for logical_ancilla_ix in [1,2]:
        if not atom_array_sim: lc.loss_noise_scale_factor = 1; lc.gate_noise_scale_factor=1
        for _ in range(QEC_cycles//2):
            lc.append_from_stim_program_text("""TICK""") # starting a QEC round
            lc.append(qec.surface_code.measure_stabilizers, [logical_ancilla_ix], order='fowler', with_cnot=biased_pres_gates, compare_with_previous=True) # append QEC rounds
            lc.append_from_stim_program_text("""TICK""") # starting a QEC round

        lc.spam_noise_scale_factor = 0  # TODO: check this. also do we need to set "lc.loss_noise_scale_factor = 0" ?
        lc.gate_noise_scale_factor = 0
        lc.loss_noise_scale_factor = 0

        if (ancilla1_for_preselection and logical_ancilla_ix == 1):
            lq = lc.logical_qubits[logical_ancilla_ix]
            physical_data_qubit_layout = lq.data_qubits.reshape(lq.dy, lq.dx)
            logical_x = physical_data_qubit_layout[lq.dy // 2, :]
            logical_x_operator = []
            for _ in range(len(logical_x)):
                logical_x_operator.append(stim.target_x(logical_x[_]))
                if _ != len(logical_x) - 1:
                    logical_x_operator.append(stim.target_combiner())
            lc.append('MPP', logical_x_operator)  # TODO: check this.
            lc.append('OBSERVABLE_INCLUDE', [stim.target_rec(-1)], lc.num_observables)
        if (ancilla2_for_preselection and logical_ancilla_ix == 2):
            lc.append('MPP', lc.logical_qubits[logical_ancilla_ix].logical_z_operator)  # TODO: check this.
            lc.append('OBSERVABLE_INCLUDE', [stim.target_rec(-1)], lc.num_observables)

        lc.spam_noise_scale_factor = 0
        lc.gate_noise_scale_factor = 1
        lc.loss_noise_scale_factor = 1

        for _ in range(QEC_cycles//2):
            lc.append_from_stim_program_text("""TICK""") # starting a QEC round
            lc.append(qec.surface_code.measure_stabilizers, [logical_ancilla_ix], order='fowler', with_cnot=biased_pres_gates, compare_with_previous=True) # append QEC rounds
            lc.append_from_stim_program_text("""TICK""") # starting a QEC round

        if not atom_array_sim: lc.loss_noise_scale_factor = 0; lc.gate_noise_scale_factor=0  # PB: what's the purpose of this line?
    
    
    if not atom_array_sim: lc.loss_noise_scale_factor = 1; lc.gate_noise_scale_factor=1
    
    
    if Steane_type == 'Regular':
        # Entangle 0 and 1:
        lc.append(qec.surface_code.global_h, [1], move_duration=0)
        lc.append(qec.surface_code.global_cz, [0,1], move_duration=0)
        lc.append(qec.surface_code.global_h, [1], move_duration=0)
        lc.append(qec.surface_code.measure_z, [1], observable_include=False)
        # Entangle 0 and 2:
        lc.append(qec.surface_code.global_h, [0], move_duration=0)
        lc.append(qec.surface_code.global_cz, [0,2], move_duration=0)
        lc.append(qec.surface_code.global_h, [0], move_duration=0)
        lc.append(qec.surface_code.measure_x, [2], observable_include=False)
    
    elif Steane_type == 'SWAP':
        # Entangle 0 and 1:
        # CX 1-->0
        lc.append(qec.surface_code.global_h, [0], move_duration=0)
        lc.append(qec.surface_code.global_cz, [0,1], move_duration=0)
        lc.append(qec.surface_code.global_h, [0], move_duration=0)
        # noiseless CX 0-->1:
        lc.loss_noise_scale_factor = 0; lc.gate_noise_scale_factor=0
        lc.append(qec.surface_code.global_h, [1], move_duration=0)
        lc.append(qec.surface_code.global_cz, [0,1], move_duration=0)
        lc.append(qec.surface_code.global_h, [1], move_duration=0)
        lc.loss_noise_scale_factor = 1; lc.gate_noise_scale_factor=1
        
        lc.append(qec.surface_code.measure_z, [0], observable_include=False)
        
        # Entangle 1 and 2 
        # CX 1-->2:
        lc.append(qec.surface_code.global_h, [2], move_duration=0)
        lc.append(qec.surface_code.global_cz, [1,2], move_duration=0)
        lc.append(qec.surface_code.global_h, [2], move_duration=0)
        
        # noiseless CX 2-->1:
        lc.loss_noise_scale_factor = 0; lc.gate_noise_scale_factor=0
        lc.append(qec.surface_code.global_h, [1], move_duration=0)
        lc.append(qec.surface_code.global_cz, [1,2], move_duration=0)
        lc.append(qec.surface_code.global_h, [1], move_duration=0)
        lc.loss_noise_scale_factor = 1; lc.gate_noise_scale_factor=1
        
        lc.append(qec.surface_code.measure_x, [1], observable_include=False)
        
    if not atom_array_sim: lc.loss_noise_scale_factor = 0; lc.gate_noise_scale_factor=0
    
    # Final measurement of the logical qubit:
    final_logical_ix = 0 if Steane_type == 'Regular' else 2
    observable_include = False if (ancilla1_for_preselection or ancilla2_for_preselection) else True
    if logical_basis == 'X':
        lc.append(qec.surface_code.measure_x, [final_logical_ix], observable_include=observable_include)

    elif logical_basis == 'Z':
        lc.append(qec.surface_code.measure_z, [final_logical_ix], observable_include=observable_include)

    return lc


def GHZ_experiment_Surface(d, order, num_logicals, code, QEC_cycles, entangling_gate_error_rate, entangling_gate_loss_rate, erasure_ratio, logical_basis='X', biased_pres_gates = False, loss_detection_on_all_qubits = True, atom_array_sim=False):
    assert logical_basis in ['X', 'Z'] # measurement basis for final logical state
    print(f"order = {order}")
    if code == 'Rotated_Surface':
        logical_qubits = [qec.surface_code.RotatedSurfaceCode(d, d) for _ in range(num_logicals)]
    elif code == 'Surface':
        logical_qubits = [qec.surface_code.SurfaceCode(d, d) for _ in range(num_logicals)]
    
    if atom_array_sim:
        lc = LogicalCircuit(logical_qubits, initialize_circuit=False,
                                loss_noise_scale_factor=1, spam_noise_scale_factor=1,
                                gate_noise_scale_factor=1, idle_noise_scale_factor=1,
                                atom_array_sim = atom_array_sim)
    else:
        lc = LogicalCircuit(logical_qubits, initialize_circuit=False,
                                loss_noise_scale_factor=1, spam_noise_scale_factor=0,
                                gate_noise_scale_factor=1, idle_noise_scale_factor=0,
                                entangling_gate_error_rate=entangling_gate_error_rate, 
                                entangling_gate_loss_rate=entangling_gate_loss_rate,
                                erasure_ratio = erasure_ratio,
                                atom_array_sim = atom_array_sim)
        
    if not atom_array_sim: lc.loss_noise_scale_factor = 0; lc.gate_noise_scale_factor=0
    
    # First prepare all logical qubits in |0>
    lc.append(qec.surface_code.prepare_zero, list(range(0, len(logical_qubits))))

    # Rotate noiselessly
    lc.append(qec.surface_code.rotate_code, 0)

    # Hadamard all qubits
    lc.append(qec.surface_code.global_h, list(range(len(logical_qubits))), move_duration=0)

    # Entangle all qubits:
    for _ in range(len(order)):
        lc.append(qec.surface_code.global_cz, [order[_][0], order[_][1]], move_duration=0)
        lc.append(qec.surface_code.global_h, order[_][1], move_duration=0)
        
    # add QEC cycles:
    if not atom_array_sim: lc.loss_noise_scale_factor = 1; lc.gate_noise_scale_factor=1
    for cycle_num in range(QEC_cycles):
        lc.append_from_stim_program_text("""TICK""") # starting a QEC round
        if loss_detection_on_all_qubits: # LD rounds on all logical qubits after every gate
            lc.append(qec.surface_code.measure_stabilizers, list(range(len(logical_qubits))), order='fowler', with_cnot=biased_pres_gates, compare_with_previous=True) # append QEC rounds
        else: # only on the logical qubits who did the gate:
            lc.append(qec.surface_code.measure_stabilizers, [order[_][0], order[_][1]], order='fowler', with_cnot=biased_pres_gates, compare_with_previous=True) # append QEC rounds
        lc.append_from_stim_program_text("""TICK""") # starting a QEC round
    if not atom_array_sim: lc.loss_noise_scale_factor = 0; lc.gate_noise_scale_factor=0
        
    if logical_basis == 'X':
        lc.append(qec.surface_code.measure_x, list(range(len(logical_qubits))), observable_include=False)
        lc.append('MOVE_TO_NO_NOISE', lc.qubit_indices, 0)
        global_x = []
        for index in range(len(lc.logical_qubits)):
            global_x += lc.logical_qubits[index].logical_x_operator + [stim.GateTarget(stim.target_combiner())]
        lc.append('MPP', global_x[:-1])
        lc.append('OBSERVABLE_INCLUDE', [stim.target_rec(-1)], 0)

    elif logical_basis == 'Z':
        lc.append(qec.surface_code.measure_z, list(range(len(logical_qubits))), observable_include=False)
        lc.append('MOVE_TO_NO_NOISE', lc.qubit_indices, 0)

        for index in range(len(logical_qubits) - 1):
            lc.append('MPP', lc.logical_qubits[index].logical_z_operator + [stim.GateTarget(stim.target_combiner())] +
                    lc.logical_qubits[index + 1].logical_z_operator)
            lc.append('OBSERVABLE_INCLUDE', [stim.target_rec(-1)], lc.num_observables)

    return lc
    
    
    

# def memory_experiment_xzzx(d, cycles, entangling_gate_error_rate, entangling_gate_loss_rate, num_logicals=1, logical_basis='X', biased_pres_gates=False):
#     assert logical_basis in ['X', 'Z']
    
#     print(f"entangling Pauli error rate = {entangling_gate_error_rate}, entangling loss rate = {entangling_gate_loss_rate}")
#     logical_qubits = [qec.codes.xzzx_code.RotatedXZZXCode(d, d) for _ in range(num_logicals)]
#     lc = LogicalCircuit(logical_qubits, initialize_circuit=False,
#                             loss_noise_scale_factor=1, spam_noise_scale_factor=0,
#                             gate_noise_scale_factor=1, idle_noise_scale_factor=0,
#                             entangling_gate_error_rate=entangling_gate_error_rate, entangling_gate_loss_rate=entangling_gate_loss_rate)
    
#     if logical_basis == 'X':
#         lc.append(qec.xzzx_code.prepare_plus, list(range(0, len(logical_qubits))), order='fowler', with_cnot=biased_pres_gates)
#         for cycle_num in range(cycles-1):
#             lc.append_from_stim_program_text("""TICK""") # starting a QEC round
#             lc.append(qec.xzzx_code.measure_stabilizers, list(range(len(logical_qubits))), order='fowler', with_cnot=biased_pres_gates, compare_with_previous=True) # append QEC rounds
#         lc.append(qec.xzzx_code.measure_z, list(range(len(logical_qubits))), observable_include=True)

#     elif logical_basis == 'Z':
#         lc.append(qec.xzzx_code.prepare_zero, list(range(0, len(logical_qubits))), order='fowler', with_cnot=biased_pres_gates)
#         for cycle_num in range(cycles-1):
#             lc.append_from_stim_program_text("""TICK""") # starting a QEC round
#             lc.append(qec.xzzx_code.measure_stabilizers, list(range(0, len(logical_qubits))), order='fowler', with_cnot=biased_pres_gates, compare_with_previous=True) # append QEC rounds
#         lc.append(qec.xzzx_code.measure_z, list(range(len(logical_qubits))), observable_include=True)

#     return lc


# def memory_experiment_surface(d, code, QEC_cycles, entangling_gate_error_rate, entangling_gate_loss_rate, num_logicals=1, logical_basis='X', biased_pres_gates = False):
#     """ This circuit simulated 1 logical qubits, a memory experiment with QEC cycles. We take perfect initialization and measurement and put noise only on the QEC cycles part."""
#     assert logical_basis in ['X', 'Z'] # init and measurement basis for the single qubit logical state
#     print(f"entangling Pauli error rate = {entangling_gate_error_rate}, entangling loss rate = {entangling_gate_loss_rate}")
    
#     if code == 'Rotated_Surface':
#         logical_qubits = [qec.surface_code.RotatedSurfaceCode(d, d) for _ in range(num_logicals)]
#     elif code == 'Surface':
#         logical_qubits = [qec.surface_code.SurfaceCode(d, d) for _ in range(num_logicals)]
        
#     lc = LogicalCircuit(logical_qubits, initialize_circuit=False,
#                             loss_noise_scale_factor=1, spam_noise_scale_factor=0,
#                             gate_noise_scale_factor=1, idle_noise_scale_factor=0,
#                             entangling_gate_error_rate=entangling_gate_error_rate, entangling_gate_loss_rate=entangling_gate_loss_rate)
#     lc.loss_noise_scale_factor = 0; lc.gate_noise_scale_factor=0
    
#     if logical_basis == 'X':
#         lc.append(qec.surface_code.prepare_plus, list(range(0, len(logical_qubits))), order='fowler', with_cnot=biased_pres_gates)
#         lc.loss_noise_scale_factor = 1; lc.gate_noise_scale_factor=1
#         for round_ix in range(QEC_cycles):
#             lc.append_from_stim_program_text("""TICK""") # starting a QEC round
#             lc.append(qec.surface_code.measure_stabilizers, list(range(len(logical_qubits))), order='fowler', with_cnot=biased_pres_gates, compare_with_previous=True) # append QEC rounds
#             lc.append_from_stim_program_text("""TICK""") # starting a QEC round
#         lc.loss_noise_scale_factor = 0; lc.gate_noise_scale_factor=0
#         lc.append(qec.surface_code.measure_x, list(range(len(logical_qubits))), observable_include=True)

#     elif logical_basis == 'Z':
#         lc.append(qec.surface_code.prepare_zero, list(range(0, len(logical_qubits))), order='fowler', with_cnot=biased_pres_gates)
#         lc.loss_noise_scale_factor = 1; lc.gate_noise_scale_factor=1
#         for round_ix in range(QEC_cycles):
#             lc.append_from_stim_program_text("""TICK""") # starting a QEC round
#             lc.append(qec.surface_code.measure_stabilizers, list(range(len(logical_qubits))), order='fowler', with_cnot=biased_pres_gates, compare_with_previous=True) # append QEC rounds
#             lc.append_from_stim_program_text("""TICK""") # starting a QEC round
#         lc.loss_noise_scale_factor = 0; lc.gate_noise_scale_factor=0
#         lc.append(qec.surface_code.measure_z, list(range(len(logical_qubits))), observable_include=True)
#     return lc

# def memory_experiment_surface_theory(d, code, QEC_cycles, entangling_gate_error_rate, entangling_gate_loss_rate, num_logicals=1, logical_basis='X', biased_pres_gates = False, loss_detection_method = 'FREE', loss_detection_frequency = 1, atom_array_sim=False):
#     """ This circuit simulated 1 logical qubits, a memory experiment with QEC cycles. We take perfect initialization and measurement and put noise only on the QEC cycles part."""
#     assert logical_basis in ['X', 'Z'] # init and measurement basis for the single qubit logical state
#     assert atom_array_sim == False
#     print(f"entangling Pauli error rate = {entangling_gate_error_rate}, entangling loss rate = {entangling_gate_loss_rate}")
    
#     if code == 'Rotated_Surface':
#         logical_qubits = [qec.surface_code.RotatedSurfaceCode(d, d) for _ in range(num_logicals)]
#     elif code == 'Surface':
#         logical_qubits = [qec.surface_code.SurfaceCode(d, d) for _ in range(num_logicals)]
    

#     lc = LogicalCircuit(logical_qubits, initialize_circuit=False,
#                             loss_noise_scale_factor=1, spam_noise_scale_factor=0,
#                             gate_noise_scale_factor=1, idle_noise_scale_factor=0,
#                             entangling_gate_error_rate=entangling_gate_error_rate, 
#                             entangling_gate_loss_rate=entangling_gate_loss_rate,
#                             atom_array_sim = atom_array_sim)

    
#     lc.loss_noise_scale_factor = 0; lc.gate_noise_scale_factor=0
    
#     if logical_basis == 'X':
#         lc.append(qec.surface_code.prepare_plus, list(range(0, len(logical_qubits))), order='fowler', with_cnot=biased_pres_gates)
#         lc.loss_noise_scale_factor = 1; lc.gate_noise_scale_factor=1
        
#         SWAP_round_index = 0; SWAP_round_type = 'none'; SWAP_round = False
#         for round_ix in range(QEC_cycles):
#             if loss_detection_method == 'SWAP' and ((round_ix+1)%loss_detection_frequency == 0): # check if its a SWAP round:
#                 SWAP_round = True
#                 SWAP_round_type = 'even' if SWAP_round_index%2 ==0 else 'odd'
#                 SWAP_round_index += 1
            
#             SWAP_round = False # cancel the SWAP operations
#             lc.append_from_stim_program_text("""TICK""") # starting a QEC round
#             lc.append(qec.surface_code.measure_stabilizers, list(range(len(logical_qubits))), order='fowler', with_cnot=biased_pres_gates, compare_with_previous=True, SWAP_round = SWAP_round, SWAP_round_type=SWAP_round_type) # append QEC rounds
#             lc.append_from_stim_program_text("""TICK""") # starting a QEC round
#         lc.loss_noise_scale_factor = 0; lc.gate_noise_scale_factor=0
#         lc.append(qec.surface_code.measure_x, list(range(len(logical_qubits))), observable_include=True)

#     elif logical_basis == 'Z':
#         lc.append(qec.surface_code.prepare_zero, list(range(0, len(logical_qubits))), order='fowler', with_cnot=biased_pres_gates)
#         lc.loss_noise_scale_factor = 1; lc.gate_noise_scale_factor=1
#         for round_ix in range(QEC_cycles):
#             lc.append_from_stim_program_text("""TICK""") # starting a QEC round
#             lc.append(qec.surface_code.measure_stabilizers, list(range(len(logical_qubits))), order='fowler', with_cnot=biased_pres_gates, compare_with_previous=True) # append QEC rounds
#             lc.append_from_stim_program_text("""TICK""") # starting a QEC round
#         lc.loss_noise_scale_factor = 0; lc.gate_noise_scale_factor=0
#         lc.append(qec.surface_code.measure_z, list(range(len(logical_qubits))), observable_include=True)
    
#     return lc


# def memory_experiment_surface_atom_array(d, code, QEC_cycles, entangling_gate_error_rate, entangling_gate_loss_rate, num_logicals=1, logical_basis='X', biased_pres_gates = False, loss_detection_method = 'FREE', loss_detection_frequency = 1, atom_array_sim=False):
#     """ This circuit simulated 1 logical qubits, a memory experiment with QEC cycles. We take perfect initialization and measurement and put noise only on the QEC cycles part."""
#     assert logical_basis in ['X', 'Z'] # init and measurement basis for the single qubit logical state
#     assert atom_array_sim == True
#     print(f"entangling Pauli error rate = {entangling_gate_error_rate}, entangling loss rate = {entangling_gate_loss_rate}")
    
#     if code == 'Rotated_Surface':
#         logical_qubits = [qec.surface_code.RotatedSurfaceCode(d, d) for _ in range(num_logicals)]
#     elif code == 'Surface':
#         logical_qubits = [qec.surface_code.SurfaceCode(d, d) for _ in range(num_logicals)]
    

#     lc = LogicalCircuit(logical_qubits, initialize_circuit=False,
#                                 loss_noise_scale_factor=1, spam_noise_scale_factor=1,
#                                 gate_noise_scale_factor=1, idle_noise_scale_factor=1,
#                                 atom_array_sim = atom_array_sim)

    
#     # lc.loss_noise_scale_factor = 0; lc.gate_noise_scale_factor=0
    
#     if logical_basis == 'X':
#         lc.append(qec.surface_code.prepare_plus, list(range(0, len(logical_qubits))), order='fowler', with_cnot=biased_pres_gates)
#         # lc.loss_noise_scale_factor = 1; lc.gate_noise_scale_factor=1
        
#         SWAP_round_index = 0; SWAP_round_type = 'none'; SWAP_round = False
#         for round_ix in range(QEC_cycles):
#             if loss_detection_method == 'SWAP' and ((round_ix+1)%loss_detection_frequency == 0): # check if its a SWAP round:
#                 SWAP_round = True
#                 SWAP_round_type = 'even' if SWAP_round_index%2 ==0 else 'odd'
#                 SWAP_round_index += 1
            
#             SWAP_round = False # cancel the SWAP operations
#             lc.append_from_stim_program_text("""TICK""") # starting a QEC round
#             lc.append(qec.surface_code.measure_stabilizers, list(range(len(logical_qubits))), order='fowler', with_cnot=biased_pres_gates, compare_with_previous=True, SWAP_round = SWAP_round, SWAP_round_type=SWAP_round_type) # append QEC rounds
#             lc.append_from_stim_program_text("""TICK""") # starting a QEC round
#         # lc.loss_noise_scale_factor = 0; lc.gate_noise_scale_factor=0
#         lc.append(qec.surface_code.measure_x, list(range(len(logical_qubits))), observable_include=True)

#     elif logical_basis == 'Z':
#         lc.append(qec.surface_code.prepare_zero, list(range(0, len(logical_qubits))), order='fowler', with_cnot=biased_pres_gates)
#         # lc.loss_noise_scale_factor = 1; lc.gate_noise_scale_factor=1
#         for round_ix in range(QEC_cycles):
#             lc.append_from_stim_program_text("""TICK""") # starting a QEC round
#             lc.append(qec.surface_code.measure_stabilizers, list(range(len(logical_qubits))), order='fowler', with_cnot=biased_pres_gates, compare_with_previous=True) # append QEC rounds
#             lc.append_from_stim_program_text("""TICK""") # starting a QEC round
#         # lc.loss_noise_scale_factor = 0; lc.gate_noise_scale_factor=0
#         lc.append(qec.surface_code.measure_z, list(range(len(logical_qubits))), observable_include=True)
    
#     return lc


if __name__ == '__main__':
    pass




import numpy as np
import qec
import pymatching, stim


def ghz(d, eta, order, num_qubits=3, measure='x'):
    assert measure in ['x', 'z']
    p = 0.001
    logical_qubits = [qec.surface_code.RotatedSurfaceCode(d, d) for _ in range(num_qubits)]
    lc = qec.LogicalCircuit(logical_qubits, initialize_circuit=False,
                            loss_noise_scale_factor=0.0, spam_noise_scale_factor=1,
                            gate_noise_scale_factor=1, idle_noise_scale_factor=0,
                            measurement_error_rate=1e-7,
                            reset_error_rate=0,
                            entangling_error_rate=(eta * p / (1+eta) / 2, eta * p / (1+eta) / 2, eta * p / (1+eta)),
                            single_qubit_error_rate=(0, 0, 0))

    # First prepare all logical qubits in |0>
    lc.append(qec.surface_code.prepare_zero, list(range(0, len(logical_qubits))))

    # Rotate noiselessly
    lc.append(qec.surface_code.rotate_code, 0)

    # Hadamard all qubits
    lc.append(qec.surface_code.global_h, list(range(len(logical_qubits))), move_duration=0)

    for _ in range(len(order)):
        lc.append(qec.surface_code.global_cz, [order[_][0], order[_][1]], move_duration=0)
        lc.append(qec.surface_code.global_h, order[_][1], move_duration=0)

    if measure == 'x':
        lc.append(qec.surface_code.measure_x, list(range(len(logical_qubits))), observable_include=False)
        lc.append('MOVE_TO_NO_NOISE', lc.qubit_indices, 0)
        global_x = []
        for index in range(len(lc.logical_qubits)):
            global_x += lc.logical_qubits[index].logical_x_operator + [stim.GateTarget(stim.target_combiner())]
        lc.append('MPP', global_x[:-1])
        lc.append('OBSERVABLE_INCLUDE', [stim.target_rec(-1)], 0)

    elif measure == 'z':
        lc.append(qec.surface_code.measure_z, list(range(len(logical_qubits))), observable_include=False)
        lc.append('MOVE_TO_NO_NOISE', lc.qubit_indices, 0)

        for index in range(len(logical_qubits) - 1):
            lc.append('MPP', lc.logical_qubits[index].logical_z_operator + [stim.GateTarget(stim.target_combiner())] +
                      lc.logical_qubits[index + 1].logical_z_operator)
            lc.append('OBSERVABLE_INCLUDE', [stim.target_rec(-1)], lc.num_observables)

    return lc



#biases = 10**np.linspace(0, 2., 10)
d = 7

z_fidelities_1 = []
x_fidelities_1 = []
z_fidelities_2 = []
x_fidelities_2 = []

for n in np.arange(2, 8, 1):
    order_1 = []
    order_2 = []
    for i in range(1, n):
        order_1.append((0, i))
        order_2.append((i - 1, i))

    orders = [order_1, order_2]
    print('GHZ state of {} logical qubits with d = {}'.format(n, d))
    for (o, order) in enumerate(orders):
        for bias in [1/2]:
            for measure in ['x', 'z']:
                num_shots = 10000
                lc = ghz(d, bias, order, num_qubits=n, measure=measure)
                dem = lc.without_loss().detector_error_model(decompose_errors=False, approximate_disjoint_errors=True)

                # Sample the circuit.
                shots = lc.compile_sampler().sample(num_shots)
                detector_shots, observable_shots = lc.without_loss().compile_m2d_converter().convert(measurements=shots,
                                                                                                     separate_observables=True)

                prediction = qec.correlated_decoders.mle.decode_gurobi_with_dem(dem, detector_shots)
                corrected_observables = 1 - np.logical_xor(observable_shots, prediction)
                if measure == 'x' and o == 0:
                    x_fidelities_1.append(np.mean(corrected_observables))
                elif measure == 'x' and o == 1:
                    x_fidelities_2.append(np.mean(corrected_observables))
                elif measure == 'z' and o == 0:
                    z_fidelities_1.append(np.mean(np.alltrue(corrected_observables == np.ones(n-1), axis=1)))
                elif measure == 'z' and o == 1:
                    z_fidelities_2.append(np.mean(np.alltrue(corrected_observables == np.ones(n-1), axis=1)))


    print(x_fidelities_1)
    print(x_fidelities_2)
    print(z_fidelities_1)
    print(z_fidelities_2)




class StabilizerCode:
    def __init__(self) -> None:
        self.stabilizers = self.get_stabilizers()

    def get_stabilizers(self):
        pass
    
    def get_data_ancilla_indices(self):
        pass
    
    def get_circuit(self):
        pass



import numpy as np
import matplotlib.pyplot as plt
from BiasedErasure.main_code.noise_channels import *
import qec
import progressbar
import time
import sinter
import math
import pickle
from BiasedErasure.main_code.circuits import *
from distutils.util import strtobool
from BiasedErasure.delayed_erasure_decoders.MLE_Loss_Decoder import MLE_Loss_Decoder
import os.path
import os
import json
from hashlib import sha256
import pickle
from BiasedErasure.main_code.XZZX import XZZX

class Simulator:
    def __init__(self, Meta_params, 
                bloch_point_params,
                noise = None,
                atom_array_sim = False,
                phys_err_vec = [],
                loss_detection_method = None,
                cycles = None,
                output_dir = None,
                save_filename = None,
                first_comb_weight=0.5,
                dont_use_loss_decoder=False
                ) -> None:
        """
        Architecture is one of 'CBQC', 'MBQC'
        Code is one of: 'Rotated_Surface', 'Surface'
        """
        self.Meta_params = Meta_params
        self.bloch_point_params = bloch_point_params
        self.architecture = Meta_params['architecture']
        self.code = Meta_params['code'] # code class in qec
        self.num_logicals = int(Meta_params['num_logicals'])
        self.bias_ratio = float(bloch_point_params['bias_ratio'])
        self.erasure_ratio = float(bloch_point_params['erasure_ratio'])
        self.bias_preserving_gates = strtobool(Meta_params['bias_preserving_gates'])
        self.logical_basis = Meta_params['logical_basis']
        self.noise = noise
        self.atom_array_sim = atom_array_sim
        self.phys_err_vec = phys_err_vec
        self.is_erasure_biased = strtobool(Meta_params['is_erasure_biased'])
        self.loss_detection_freq = int(Meta_params['LD_freq'])
        self.SSR = strtobool(Meta_params['SSR'])
        self.heralded_circuit = loss_detection_method
        self.cycles = cycles
        self.cycles_str = Meta_params['cycles']
        self.loss_detection_method_str = Meta_params['LD_method']
        self.ordering_type = Meta_params['ordering'] # relevant for rotated codes
        self.loss_decoder = Meta_params['loss_decoder']
        self.decoder = Meta_params['decoder']
        self.loss_decoder_type = Meta_params['loss_decoder']
        self.circuit_type = Meta_params['circuit_type']
        self.Steane_type = Meta_params['Steane_type']
        self.printing = strtobool(Meta_params['printing'])
        self.output_dir = output_dir
        self.save_filename = save_filename
        self.first_comb_weight = first_comb_weight
        self.dont_use_loss_decoder = dont_use_loss_decoder # if TRUE, we are not using loss decoder at all. all shots get same DEM.

    def get_job_id(self):
        # Check for environment variables used by different cluster management systems
        for env_var in ['SLURM_JOB_ID', 'PBS_JOBID', 'LSB_JOBID']:
            job_id = os.environ.get(env_var)
            if job_id is not None:
                return job_id
        return None  # or raise an error if the job ID is critical

    def generate_unique_key(self):
        # Convert the data to a JSON string
        data_str = json.dumps({
            'meta_params': self.Meta_params,
            'distance': self.distance
        }, sort_keys=True)
        # Use SHA256 to generate a unique hash of the data
        return sha256(data_str.encode()).hexdigest()
    

    def generate_circuit(self, d, cycles, phys_err):
        entangling_gate_error_rate, entangling_gate_loss_rate = self.noise(phys_err, self.bias_ratio)
        if self.circuit_type == 'memory':
            if self.architecture == 'CBQC':
                return memory_experiment_surface_new(d=d, code=self.code, QEC_cycles=cycles, entangling_gate_error_rate=entangling_gate_error_rate, 
                                                entangling_gate_loss_rate=entangling_gate_loss_rate, erasure_ratio = self.erasure_ratio,
                                                num_logicals=self.num_logicals, logical_basis=self.logical_basis, 
                                                biased_pres_gates = self.bias_preserving_gates, ordering = self.ordering_type,
                                                loss_detection_method = self.loss_detection_method_str, 
                                                loss_detection_frequency = self.loss_detection_freq, atom_array_sim=self.atom_array_sim)
                
                # return memory_experiment_surface(d=d, code=self.code, QEC_cycles=cycles-1, entangling_gate_error_rate=entangling_gate_error_rate, 
                #                                 entangling_gate_loss_rate=entangling_gate_loss_rate, erasure_ratio = self.erasure_ratio,
                #                                 num_logicals=self.num_logicals, logical_basis=self.logical_basis, 
                #                                 biased_pres_gates = self.bias_preserving_gates, ordering = self.ordering_type,
                #                                 loss_detection_method = self.loss_detection_method_str, 
                #                                 loss_detection_frequency = self.loss_detection_freq, atom_array_sim=self.atom_array_sim)
            elif self.architecture == 'MBQC':
                return memory_experiment_MBQC(d=d, QEC_cycles=cycles, entangling_gate_error_rate=entangling_gate_error_rate, 
                                                entangling_gate_loss_rate=entangling_gate_loss_rate, erasure_ratio = self.erasure_ratio,
                                                logical_basis=self.logical_basis, 
                                                biased_pres_gates = self.bias_preserving_gates, atom_array_sim=self.atom_array_sim)
                
                
        elif self.circuit_type in ['GHZ_all_o1', 'GHZ_save_o1','GHZ_all_o2', 'GHZ_save_o2']:
            order_1 = []
            order_2 = []
            for i in range(1, self.num_logicals):
                order_1.append((0, i))
                order_2.append((i - 1, i))
            chosen_order = order_1 if self.circuit_type.endswith("1") else order_2
            if self.circuit_type in ['GHZ_all_o1', 'GHZ_all_o2']:
                return GHZ_experiment_Surface(d=d, order=chosen_order, num_logicals=self.num_logicals, code=self.code, QEC_cycles=cycles, 
                                            entangling_gate_error_rate=entangling_gate_error_rate, entangling_gate_loss_rate=entangling_gate_loss_rate, 
                                            erasure_ratio = self.erasure_ratio,
                                            logical_basis=self.logical_basis, biased_pres_gates = self.bias_preserving_gates, 
                                            loss_detection_on_all_qubits=True, atom_array_sim=self.atom_array_sim)
            elif self.circuit_type in ['GHZ_save_o1', 'GHZ_save_o2']:
                return GHZ_experiment_Surface(d=d, order=chosen_order, num_logicals=self.num_logicals, code=self.code, QEC_cycles=cycles, 
                                            entangling_gate_error_rate=entangling_gate_error_rate, entangling_gate_loss_rate=entangling_gate_loss_rate, 
                                            erasure_ratio = self.erasure_ratio,
                                            logical_basis=self.logical_basis, biased_pres_gates = self.bias_preserving_gates, 
                                            loss_detection_on_all_qubits=False, atom_array_sim=self.atom_array_sim)
        elif self.circuit_type == 'Steane_QEC':
            return Steane_QEC_circuit(d=d, code=self.code, Steane_type=self.Steane_type, QEC_cycles=cycles-1,
                                        entangling_gate_error_rate=entangling_gate_error_rate, entangling_gate_loss_rate=entangling_gate_loss_rate,
                                        erasure_ratio=self.erasure_ratio,
                                        logical_basis=self.logical_basis, biased_pres_gates = self.bias_preserving_gates,
                                        loss_detection_on_all_qubits=True, atom_array_sim=self.atom_array_sim)
                

        else:
            return None
    
    
    def simulate(self, distances, num_shots):
        f = open(f'{self.output_dir}/{self.save_filename}.txt', "a")
        
        for d in distances:
            cycles = d if self.cycles == None else self.cycles
                        
            for phys_err in self.phys_err_vec:
                start_time = time.time()
                
                LogicalCircuit = self.generate_circuit(d=d, cycles=cycles, phys_err=phys_err)
                if self.printing:
                    print(f"\nCircuit after noise:\n{LogicalCircuit} \n")
                    print(f"potential lost qubits: {LogicalCircuit.potential_lost_qubits} \n with loss probabilities: {LogicalCircuit.loss_probabilities}")
                
                
                if self.circuit_type == 'Steane_QEC':
                    ValueError (self.num_logicals == 3)
                    corrections, observables, probabilities1, probabilities2 = self.count_logical_errors_preselection(num_shots=num_shots, distance=d, phys_error=phys_err, cycles=cycles)
                    
                    job_id = self.get_job_id()
                    full_filename = f'{self.output_dir}/{self.save_filename}/d{d}__c{cycles}__p{phys_err}__Steane_QEC_results.pickle'

                    # Create folder if it doesn't exist.
                    folder_path = f'{self.output_dir}/{self.save_filename}'
                    if not os.path.exists(folder_path):
                        # If the folder does not exist, create it
                        os.makedirs(folder_path)
                    
                    # Data structure to append (now as a list)
                    data_to_append = [job_id, corrections, observables, probabilities1, probabilities2]
                    
                    # Open the file in append mode, binary
                    with open(full_filename, 'ab') as file:
                        pickle.dump(data_to_append, file)
                        
                    f.write(f'{d} {phys_err} {cycles} {job_id} {num_shots} {time.time()-start_time}\n')
                    
                    
                else:
                    num_errors_sampled, num_shots = self.count_logical_errors(LogicalCircuit=LogicalCircuit, num_shots=num_shots, distance=d, phys_error=phys_err, cycles=cycles)
                    print(f"for d = {d}, {cycles} cycles, physical error rate = {phys_err}, {num_shots} shots, we had {num_errors_sampled} errors (logical error = {(num_errors_sampled/num_shots):.1e})")
                    f.write(f'{d} {phys_err} {num_errors_sampled} {num_shots} {time.time()-start_time}\n')
                    
        f.close()

    
    def count_logical_errors(self, LogicalCircuit, num_shots: int, distance=None, phys_error=0, debugging=False, cycles=None):
        """This function decodes the loss information using mle. 
        Given heralded losses upon measurements, there are multiple potential loss events (with some probability) in the circuit.
        There are 2 options:
        1. MLE approximate decoding - for each potential loss event we create a DEM and connect all to generate the final DEM. We use MLE to decode given the ginal DEM and the experimental measurements.
        2. Accurate MLE - for each potential loss event i we decode with Gorubi to get P_i, and take argmax(P_i) to decode.
        """
        # decoder_type = 'independent'
        
        if self.printing:
            print(f"Starting the decoding!")
            start_time = time.time()

        if self.architecture == "MBQC":
            xzzx_instance = XZZX()
            data_qubits, ancilla_qubits = xzzx_instance.get_data_ancilla_indices(dx=distance, dy=distance, cycles=cycles, architecture="MBQC")
        else:
            ancilla_qubits = [qubit for i in range(self.num_logicals) for qubit in LogicalCircuit.logical_qubits[i].measure_qubits]
            data_qubits = [qubit for i in range(self.num_logicals) for qubit in LogicalCircuit.logical_qubits[i].data_qubits]
        
        
        sampler = LogicalCircuit.compile_detector_sampler()
        loss_sampling = 1 if self.erasure_ratio > 0 else  num_shots # how many times we will use the same loss sampling result
        num_loss_shots = math.ceil(num_shots / loss_sampling)
        num_shots = num_loss_shots * loss_sampling
        loss_detection_events_all_shots = np.random.rand(num_loss_shots, len(LogicalCircuit.potential_lost_qubits)) < LogicalCircuit.loss_probabilities
        # LogicalCircuit.logical_qubits[0].visualize_code()
        # print(LogicalCircuit.loss_probabilities)
        loss_detection_class = self.heralded_circuit(circuit=LogicalCircuit, biased_erasure=self.is_erasure_biased, bias_preserving_gates=self.bias_preserving_gates,
                                                                    basis = self.logical_basis, distance=distance, erasure_ratio = self.erasure_ratio, 
                                                                    phys_error = phys_error, ancilla_qubits=ancilla_qubits, data_qubits=data_qubits,
                                                                    SSR=self.SSR, cycles=cycles, printing=self.printing, loss_detection_freq = self.loss_detection_freq)
        
        
        MLE_Loss_Decoder_class = MLE_Loss_Decoder(Meta_params=self.Meta_params, bloch_point_params=self.bloch_point_params, distance = distance, loss_detection_method_str=self.loss_detection_method_str,
                                                ancilla_qubits=ancilla_qubits, data_qubits=data_qubits,
                                                cycles=cycles, printing=self.printing, loss_detection_freq = self.loss_detection_freq, 
                                                first_comb_weight=self.first_comb_weight,
                                                output_dir = self.output_dir, decoder_type = self.loss_decoder_type)
        
        if self.loss_detection_method_str == 'SWAP':
            SWAP_circuit = loss_detection_class.transfer_circuit_into_SWAP_circuit(LogicalCircuit)
            if self.printing:
                print(f"Logical circuit after implementing SWAP is: \n {SWAP_circuit}\n")
            # loss_detection_class.SWAP_circuit = SWAP_circuit
            MLE_Loss_Decoder_class.circuit = SWAP_circuit
        
        elif self.loss_detection_method_str in ['FREE', 'MBQC']:
            MLE_Loss_Decoder_class.circuit = LogicalCircuit
            
        
        
        if self.printing:
            end_time = time.time()
            print(f"Initialization of loss decoders (and building SWAP circuit if needed) took {end_time - start_time} sec.")
            start_time = time.time()
                
        if self.erasure_ratio > 0:
            
            MLE_Loss_Decoder_class.initialize_loss_decoder()
            
            if self.printing:
                end_time = time.time()
                print(f"Building the Pauli DEM and loading all independent loss DEMs took {end_time - start_time} sec. Starting the loss decoding!")
                start_time = time.time()
            
            num_errors = 0
            dems_list = []
            detection_events_list = []
            observable_flips_list = []
            start_time_all_shots = time.time()
            for shot in range(num_loss_shots):
                loss_detection_events = loss_detection_events_all_shots[shot]
                
                experimental_circuit, detector_error_model = MLE_Loss_Decoder_class.decode_loss_MLE(loss_detection_events)

                if self.printing:
                    # print(f"\n Loss detection events: {loss_detection_events}")
                    # print(f"\n Potential lost qubits: {LogicalCircuit.potential_lost_qubits} \n with loss probabilities: {LogicalCircuit.loss_probabilities}")
                    print("\n Experimental circuit (for measurements):")
                    print(experimental_circuit)
                    print("\n MLE DEM:")
                    print(detector_error_model)
                
                # Sample MEASUREMENTS from experimental_circuit
                sampler = experimental_circuit.compile_detector_sampler()
                detection_events, observable_flips = sampler.sample(loss_sampling, separate_observables=True)
                
                detection_events_list.append(detection_events)
                observable_flips_list.append(observable_flips)
                
                # Extract decoder configuration data from the circuit.
                if self.dont_use_loss_decoder:
                    no_loss_decoder_dem = MLE_Loss_Decoder_class.circuit.detector_error_model(decompose_errors=False, approximate_disjoint_errors=True, ignore_decomposition_failures=True, allow_gauge_detectors=False)
                    dems_list.append(no_loss_decoder_dem)
                else:
                    dems_list.append(detector_error_model)
                
            if self.printing:
                end_time_all_shots = time.time()
                print(f"Building the MLE loss DEM and experimental circuits for all shots took {end_time_all_shots - start_time_all_shots} sec")
                start_time = time.time()
            
            if self.decoder == "MLE":
                
                detector_shots = np.array(detection_events_list)
                predictions = qec.correlated_decoders.mle_loss.decode_gurobi_with_dem_loss(dems_list=dems_list, detector_shots = detector_shots)   
                if self.printing:
                    end_time = time.time()
                    print(f"Gurobi correlated mle decoding for all shots took {end_time - start_time} sec")
            else:
                predictions = []
                for (d, detection_events) in enumerate(detection_events_list):
                    detector_error_model = dems_list[d]
                    matching = pymatching.Matching.from_detector_error_model(detector_error_model)
                    prediction = matching.decode_batch(detection_events)
                    predictions.append(prediction[0][0])
                predictions = np.array(predictions)
                if self.printing:
                    end_time = time.time()
                    print(f"MWPM decoding for all shots took {end_time - start_time} sec")
                
            observable_flips = np.array(observable_flips_list)
            predictions_bool = predictions.astype(bool).squeeze()
            observable_flips_squeezed = observable_flips.squeeze()

            num_errors = np.sum(np.logical_xor(observable_flips_squeezed, predictions_bool))


                                                    
        else: # no loss errors at all
            sampler = MLE_Loss_Decoder_class.circuit.compile_detector_sampler()
            detection_events, observable_flips = sampler.sample(num_shots, separate_observables=True)
            

            if self.decoder == "MLE":
                detector_error_model = MLE_Loss_Decoder_class.circuit.detector_error_model(decompose_errors=False, approximate_disjoint_errors=True, ignore_decomposition_failures=True, allow_gauge_detectors=False)
                prediction = qec.correlated_decoders.mle.decode_gurobi_with_dem(dem=detector_error_model, detector_shots = detection_events)
            else:
                detector_error_model = MLE_Loss_Decoder_class.circuit.detector_error_model(decompose_errors=True, approximate_disjoint_errors=True, ignore_decomposition_failures=True) 
                prediction = sinter.predict_observables(
                    dem=detector_error_model,
                    dets=detection_events,
                    decoder='pymatching',
                )

            num_errors = np.sum(np.logical_xor(observable_flips, prediction))
        
        # logical_error = num_errors / num_shots
        return num_errors, num_shots
    
    
    
    def count_logical_errors_experiment(self, num_shots: int, distance: int, measurement_events: np.ndarray, detection_events_signs: np.ndarray, use_loss_decoding=True):
        """This function decodes the loss information using mle. 
        Given heralded losses upon measurements, there are multiple potential loss events (with some probability) in the circuit.
        We use the MLE approximate decoding - for each potential loss event we create a DEM and connect all to generate the final DEM. We use MLE to decode given the ginal DEM and the experimental measurements.
        Input: Meta_params, distance, num shots, experimental data: detector shots.
        Output: final DEM, corrections, num errors.
        
        Meta_params = {'architecture': 'CBQC', 'code': 'Rotated_Surface', 'logical_basis': 'X', 'bias_preserving_gates': 'False', 
                'noise': 'atom_array', 'is_erasure_biased': 'False', 'LD_freq': '1', 'LD_method': 'SWAP', 'SSR': 'True', 'cycles': '2', 'ordering': 'bad', 'decoder': 'MLE',
                'circuit_type': 'memory', 'printing': 'False', 'num_logicals': '1'}
        ordering: bad or fowler (good)
        decoder: MLE or MWPM
        """
        
        # Step 1 - generate the experimental circuit in our simulation:
        LogicalCircuit = self.generate_circuit(d=distance, cycles=self.cycles, phys_err=None)
        
        ancilla_qubits = [qubit for i in range(self.num_logicals) for qubit in LogicalCircuit.logical_qubits[i].measure_qubits]
        data_qubits = [qubit for i in range(self.num_logicals) for qubit in LogicalCircuit.logical_qubits[i].data_qubits]
                
        
        # LogicalCircuit.logical_qubits[0].visualize_code()
        loss_detection_class = self.heralded_circuit(circuit=LogicalCircuit, biased_erasure=self.is_erasure_biased, bias_preserving_gates=self.bias_preserving_gates,
                                                                    basis = self.logical_basis, distance=distance, erasure_ratio = self.erasure_ratio, 
                                                                    phys_error = None, ancilla_qubits=ancilla_qubits, data_qubits=data_qubits,
                                                                    SSR=self.SSR, cycles=self.cycles, printing=self.printing, loss_detection_freq = self.loss_detection_freq)
        
        
        MLE_Loss_Decoder_class = MLE_Loss_Decoder(Meta_params=self.Meta_params, bloch_point_params=self.bloch_point_params, distance = distance, loss_detection_method_str=self.loss_detection_method_str,
                                                ancilla_qubits=ancilla_qubits, data_qubits=data_qubits,
                                                cycles=self.cycles, printing=self.printing, loss_detection_freq = self.loss_detection_freq,
                                                output_dir = self.output_dir, decoder_type=self.loss_decoder_type)
        
        if self.loss_detection_method_str == 'SWAP':
            SWAP_circuit = loss_detection_class.transfer_circuit_into_SWAP_circuit(LogicalCircuit)
            if self.printing:
                print(f"Logical circuit after implementing SWAP is: \n {SWAP_circuit}\n")
            loss_detection_class.SWAP_circuit = SWAP_circuit
            MLE_Loss_Decoder_class.circuit = SWAP_circuit
        
        elif self.loss_detection_method_str == 'FREE':
            MLE_Loss_Decoder_class.circuit = LogicalCircuit
        
        if self.printing:
            print(f"Logical circuit that will be used: \n{MLE_Loss_Decoder_class.circuit}")
            
        if 2 in measurement_events and use_loss_decoding:
            loss_sampling = 1 # how many times we will use the same loss sampling result
            num_loss_shots = math.ceil(num_shots / loss_sampling)
            num_shots = num_loss_shots * loss_sampling
        
            MLE_Loss_Decoder_class.initialize_loss_decoder() # this part can be improved to be a bit faster
            
            # Loss decoding - creating DEMs:
            dems_list = []
            probs_lists = []
            observables_errors_interactions_lists = []
            for shot in range(num_loss_shots):
                measurement_event = measurement_events[shot] # change it to measurements
                final_dem_hyperedges_matrix, observables_errors_interactions = MLE_Loss_Decoder_class.generate_dem_loss_mle_experiment(measurement_event) # final_dem_hyperedges_matrix doesn't contain observables, only detectors               
                observables_errors_interactions_lists.append(observables_errors_interactions)
                
                if self.decoder == "MLE":
                    final_dem_hyperedges_matrix, probs_list = MLE_Loss_Decoder_class.convert_hyperedge_matrix_into_binary(hyperedges_matrix = final_dem_hyperedges_matrix)
                    dems_list.append(final_dem_hyperedges_matrix)
                    probs_lists.append(probs_list)
                else:
                    final_dem = MLE_Loss_Decoder_class.from_hyperedges_matrix_into_stim_dem(final_dem_hyperedges_matrix) # convert into stim format. #TODO: bug - fix it! here final_dem_hyperedges_matrix doesn't contain observables so the DEM will not contain them.
                    dems_list.append(final_dem)
                    
                if self.printing:
                    print("\n MLE DEM hyperedges matrix:")
                    print(final_dem_hyperedges_matrix)

            measurement_events[measurement_events == 2] = 0 #change all values in detection_events from 2 to 0
            measurement_events = measurement_events.astype(np.bool_)
            detection_events, observable_flips = MLE_Loss_Decoder_class.circuit.compile_m2d_converter().convert(measurements=measurement_events, separate_observables=True)
            
            # add normalization step of detection events:
            detection_events = detection_events * detection_events_signs
            
            # Creating the predictions using the DEM:
            if self.decoder == "MLE":
                predictions = qec.correlated_decoders.mle_loss.decode_gurobi_with_dem_loss(dems_list=dems_list, probs_lists = probs_lists, detector_shots = detection_events, observables_errors_interactions_lists=observables_errors_interactions_lists)   
                
            else:
                predictions = []
                for (d, detection_event) in enumerate(detection_events):
                    detector_error_model = dems_list[d]
                    matching = pymatching.Matching.from_detector_error_model(detector_error_model)
                    prediction = matching.decode_batch(detection_event)
                    predictions.append(prediction[0][0])
            
            num_errors = np.sum(np.logical_xor(observable_flips, predictions))
            if self.printing:
                print(f"for d = {distance}, {self.cycles} cycles, {num_shots} shots, we had {num_errors} errors (logical error = {(num_errors/num_shots):.1e})")
            predictions_bool = predictions.astype(bool).squeeze()
            return predictions_bool, dems_list
        
        
        
        
        else: # regular decoding, without delayed erasure decoder
            measurement_events[measurement_events == 2] = 0 #change all values in detection_events from 2 to 0
            measurement_events = measurement_events.astype(np.bool_)
            detection_events, observable_flips = MLE_Loss_Decoder_class.circuit.compile_m2d_converter().convert(measurements=measurement_events, separate_observables=True)
            
            # add normalization step of detection events:
            detection_events_int = detection_events.astype(np.int32)
            detection_events_flipped = np.where(detection_events_signs == -1, ~detection_events_int, detection_events_int)
            detection_events = detection_events_flipped.astype(np.bool_)

            
            
            if self.decoder == "MLE":
                detector_error_model = MLE_Loss_Decoder_class.circuit.detector_error_model(decompose_errors=False, approximate_disjoint_errors=True, ignore_decomposition_failures=True, allow_gauge_detectors=False)
                predictions = qec.correlated_decoders.mle.decode_gurobi_with_dem(dem=detector_error_model, detector_shots = detection_events)
            else:
                detector_error_model = MLE_Loss_Decoder_class.circuit.detector_error_model(decompose_errors=True, approximate_disjoint_errors=True, ignore_decomposition_failures=True) 
                predictions = sinter.predict_observables(
                    dem=detector_error_model,
                    dets=detection_events,
                    decoder='pymatching',
                )

            num_errors = np.sum(np.logical_xor(observable_flips, predictions))
            if self.printing:
                print(f"for d = {distance}, {self.cycles} cycles, {num_shots} shots, we had {num_errors} errors (logical error = {(num_errors/num_shots):.1e})")
            predictions_bool = predictions.astype(bool).squeeze()
            return predictions_bool, detector_error_model
        
    
    
    def count_logical_errors_preselection_old(self, num_shots: int, distance=None, phys_error=0, debugging=False, cycles=None):
        decoder_type = 'independent'
        d = distance

        entangling_gate_error_rate, entangling_gate_loss_rate = self.noise(phys_error, self.bias_ratio, self.erasure_ratio)
        def steane(d, ancilla1_for_preselection=False, ancilla2_for_preselection=False):
            return Steane_QEC_circuit(d=d, code=self.code, Steane_type=self.Steane_type, QEC_cycles=cycles, entangling_gate_error_rate=entangling_gate_error_rate,
                                    entangling_gate_loss_rate=entangling_gate_loss_rate, erasure_ratio=self.erasure_ratio, logical_basis=self.logical_basis, biased_pres_gates = self.bias_preserving_gates,
                                    loss_detection_on_all_qubits=True, atom_array_sim=self.atom_array_sim, ancilla1_for_preselection=ancilla1_for_preselection, ancilla2_for_preselection=ancilla2_for_preselection)

        lc = steane(d)
        
        ancilla_qubits = [qubit for i in range(self.num_logicals) for qubit in lc.logical_qubits[i].measure_qubits]
        data_qubits = [qubit for i in range(self.num_logicals) for qubit in lc.logical_qubits[i].data_qubits]
        
        loss_sampling = 1 if self.erasure_ratio > 0 else  num_shots # how many times we will use the same loss sampling result
        num_loss_shots = math.ceil(num_shots / loss_sampling)
        num_shots = num_loss_shots * loss_sampling
        loss_detection_events_all_shots = np.random.rand(num_loss_shots, len(lc.potential_lost_qubits)) < lc.loss_probabilities

        
        # get SWAP circuit for each circuit ('regular','1','2'):
        loss_detection_class = self.heralded_circuit(circuit=lc, biased_erasure=self.is_erasure_biased, bias_preserving_gates=self.bias_preserving_gates,
                                                                    basis = self.logical_basis, distance=distance, erasure_ratio = self.erasure_ratio, 
                                                                    phys_error = phys_error, ancilla_qubits=ancilla_qubits, data_qubits=data_qubits,
                                                                    SSR=self.SSR, cycles=cycles, printing=self.printing, loss_detection_freq = self.loss_detection_freq)
        SWAP_circuit_regular = loss_detection_class.transfer_circuit_into_SWAP_circuit(lc)

            
        lc_for_preselection1 = steane(d, ancilla1_for_preselection=True)
        loss_detection_class.logical_circuit = lc_for_preselection1
        SWAP_circuit_preselection1 = loss_detection_class.transfer_circuit_into_SWAP_circuit(lc_for_preselection1)

        lc_for_preselection2 = steane(d, ancilla2_for_preselection=True)
        loss_detection_class.logical_circuit = lc_for_preselection2
        SWAP_circuit_preselection2 = loss_detection_class.transfer_circuit_into_SWAP_circuit(lc_for_preselection2)


        # get DEMs for each circuit:
        MLE_Loss_Decoder_class = MLE_Loss_Decoder(Meta_params=self.Meta_params, bloch_point_params=self.bloch_point_params, distance = distance, loss_detection_method_str=self.loss_detection_method_str,
                                                ancilla_qubits=ancilla_qubits, data_qubits=data_qubits,
                                                cycles=cycles, printing=self.printing, loss_detection_freq = self.loss_detection_freq, 
                                                first_comb_weight=self.first_comb_weight,
                                                output_dir = self.output_dir, decoder_type = decoder_type)
        #MLE_Loss_Decoder_class.initialize_loss_decoder()
        SWAP_circuits = {'regular': SWAP_circuit_regular, '1': SWAP_circuit_preselection1, '2': SWAP_circuit_preselection2}
        #DEMs = {'regular': [], '1': [], '2': []}

        corrections = []
        observables = []
        probabilities1 = np.zeros((0, 2))
        probabilities2 = np.zeros((0, 2))
        for circuit_name in ['regular','1','2']:
            print(circuit_name)
            MLE_Loss_Decoder_class.circuit = SWAP_circuits[circuit_name]
            MLE_Loss_Decoder_class.initialize_loss_decoder()
            for shot in range(num_loss_shots):
                loss_detection_events = loss_detection_events_all_shots[shot]
                experimental_circuit, dem = MLE_Loss_Decoder_class.decode_loss_MLE(loss_detection_events)
                if circuit_name == 'regular':
                    # Sample MEASUREMENTS from experimental_circuit
                    sampler = experimental_circuit.compile_detector_sampler()
                    detection_events, observable_flips = sampler.sample(loss_sampling, separate_observables=True)
                    observables.extend(observable_flips)
                    corrections.extend(qec.correlated_decoders.mle.decode_gurobi_with_dem(dem, detection_events))
                elif circuit_name == '1':
                    _, prob1 = qec.correlated_decoders.mle_sliding_scale.sliding_scale(dem, detection_events[:, :dem.num_detectors])
                    probabilities1 = np.concatenate((probabilities1, prob1), axis=0)
                elif circuit_name == '2':
                    _, prob2 = qec.correlated_decoders.mle_sliding_scale.sliding_scale(dem, detection_events[:, :dem.num_detectors])
                    probabilities2 = np.concatenate((probabilities2, prob2), axis=0)
                else:
                    assert True is False

            
        return corrections, observables, probabilities1, probabilities2


    def count_logical_errors_preselection(self, num_shots: int, distance=None, phys_error=0, debugging=False, cycles=None):
        decoder_type = 'independent'
        d = distance

        entangling_gate_error_rate, entangling_gate_loss_rate = self.noise(phys_error, self.bias_ratio)
        def steane(d, ancilla1_for_preselection=False, ancilla2_for_preselection=False):
            return Steane_QEC_circuit(d=d, code=self.code, Steane_type=self.Steane_type, QEC_cycles=cycles, entangling_gate_error_rate=entangling_gate_error_rate,
                                    entangling_gate_loss_rate=entangling_gate_loss_rate, erasure_ratio=self.erasure_ratio, logical_basis=self.logical_basis, biased_pres_gates = self.bias_preserving_gates,
                                    loss_detection_on_all_qubits=True, atom_array_sim=self.atom_array_sim, ancilla1_for_preselection=ancilla1_for_preselection, ancilla2_for_preselection=ancilla2_for_preselection)

        lc = steane(d)
        
        ancilla_qubits = [qubit for i in range(self.num_logicals) for qubit in lc.logical_qubits[i].measure_qubits]
        data_qubits = [qubit for i in range(self.num_logicals) for qubit in lc.logical_qubits[i].data_qubits]
        
        loss_sampling = 1 if self.erasure_ratio > 0 else  num_shots # how many times we will use the same loss sampling result
        num_loss_shots = math.ceil(num_shots / loss_sampling)
        num_shots = num_loss_shots * loss_sampling
        loss_detection_events_all_shots = np.random.rand(num_loss_shots, len(lc.potential_lost_qubits)) < lc.loss_probabilities

        
        # get SWAP circuit for each circuit ('regular','1','2'):
        loss_detection_class = self.heralded_circuit(circuit=lc, biased_erasure=self.is_erasure_biased, bias_preserving_gates=self.bias_preserving_gates,
                                                                    basis = self.logical_basis, distance=distance, erasure_ratio = self.erasure_ratio, 
                                                                    phys_error = phys_error, ancilla_qubits=ancilla_qubits, data_qubits=data_qubits,
                                                                    SSR=self.SSR, cycles=cycles, printing=self.printing, loss_detection_freq = self.loss_detection_freq)
        if self.loss_detection_method_str == 'SWAP':
            SWAP_circuit_regular = loss_detection_class.transfer_circuit_into_SWAP_circuit(lc)
        else:
            SWAP_circuit_regular = lc
            
        lc_for_preselection1 = steane(d, ancilla1_for_preselection=True)
        loss_detection_class.logical_circuit = lc_for_preselection1
        if self.loss_detection_method_str == 'SWAP':
            SWAP_circuit_preselection1 = loss_detection_class.transfer_circuit_into_SWAP_circuit(lc_for_preselection1)
        else:
            SWAP_circuit_preselection1 = lc_for_preselection1

        lc_for_preselection2 = steane(d, ancilla2_for_preselection=True)
        loss_detection_class.logical_circuit = lc_for_preselection2
        if self.loss_detection_method_str == 'SWAP':
            SWAP_circuit_preselection2 = loss_detection_class.transfer_circuit_into_SWAP_circuit(lc_for_preselection2)
        else:
            SWAP_circuit_preselection2 = lc_for_preselection2

        # get DEMs for each circuit:
        MLE_Loss_Decoder_class = MLE_Loss_Decoder(Meta_params=self.Meta_params, bloch_point_params=self.bloch_point_params, distance = distance, loss_detection_method_str=self.loss_detection_method_str,
                                                ancilla_qubits=ancilla_qubits, data_qubits=data_qubits,
                                                cycles=cycles, printing=self.printing, loss_detection_freq = self.loss_detection_freq, 
                                                first_comb_weight=self.first_comb_weight,
                                                output_dir = self.output_dir, decoder_type = decoder_type)
        #MLE_Loss_Decoder_class.initialize_loss_decoder()
        SWAP_circuits = {'regular': SWAP_circuit_regular, '1': SWAP_circuit_preselection1, '2': SWAP_circuit_preselection2}
        #DEMs = {'regular': [], '1': [], '2': []}

        corrections = []
        observables = []
        probabilities1 = np.zeros((0, 2))
        probabilities2 = np.zeros((0, 2))
        for circuit_name in ['regular','1','2']:
            print(circuit_name)
            
            if self.erasure_ratio > 0:
                MLE_Loss_Decoder_class.circuit = SWAP_circuits[circuit_name]
                MLE_Loss_Decoder_class.initialize_loss_decoder()
                for shot in range(num_loss_shots):
                    loss_detection_events = loss_detection_events_all_shots[shot]
                    experimental_circuit, dem = MLE_Loss_Decoder_class.decode_loss_MLE(loss_detection_events)
                    if circuit_name == 'regular':
                        # Sample MEASUREMENTS from experimental_circuit
                        sampler = experimental_circuit.compile_detector_sampler()
                        detection_events, observable_flips = sampler.sample(loss_sampling, separate_observables=True)
                        observables.extend(observable_flips)
                        corrections.extend(qec.correlated_decoders.mle.decode_gurobi_with_dem(dem, detection_events))
                    elif circuit_name == '1':
                        # _, prob1 = qec.correlated_decoders.mle_sliding_scale.sliding_scale(dem, detection_events[:, :dem.num_detectors])
                        _, prob1 = qec.correlated_decoders.mle_sliding_scale.sliding_scale(dem, detection_events)
                        probabilities1 = np.concatenate((probabilities1, prob1), axis=0)
                    elif circuit_name == '2':
                        # _, prob2 = qec.correlated_decoders.mle_sliding_scale.sliding_scale(dem, detection_events[:, :dem.num_detectors])
                        _, prob2 = qec.correlated_decoders.mle_sliding_scale.sliding_scale(dem, detection_events)
                        probabilities2 = np.concatenate((probabilities2, prob2), axis=0)
                    else:
                        assert True is False
                        
            else: # loss ratio = 0, execute all shot together
                dem = SWAP_circuits[circuit_name].detector_error_model(decompose_errors=False, approximate_disjoint_errors=True, ignore_decomposition_failures=True, allow_gauge_detectors=False)
                if circuit_name == 'regular':
                    sampler = SWAP_circuits[circuit_name].compile_detector_sampler()
                    detection_events, observable_flips = sampler.sample(num_shots, separate_observables=True)
                    observables.extend(observable_flips)
                    corrections.extend(qec.correlated_decoders.mle.decode_gurobi_with_dem(dem, detection_events))
                elif circuit_name == '1':
                    _, prob1 = qec.correlated_decoders.mle_sliding_scale.sliding_scale(dem, detection_events[:, :dem.num_detectors])
                    probabilities1 = np.concatenate((probabilities1, prob1), axis=0)
                elif circuit_name == '2':
                    _, prob2 = qec.correlated_decoders.mle_sliding_scale.sliding_scale(dem, detection_events[:, :dem.num_detectors])
                    probabilities2 = np.concatenate((probabilities2, prob2), axis=0)
                else:
                    assert True is False
            
        return corrections, observables, probabilities1, probabilities2



from setuptools import setup, find_packages

setup(
    name='BiasedErasure',
    version='0.1.6',
    author='Gefen Baranes and Pablo Bonilla',
    author_email='gefenbaranes123@gmail.com',
    packages=find_packages(include=['BiasedErasure', 'BiasedErasure.*']),
    license='LICENSE',
    description='Simulate logical circuits with loss errors with efficient smart decoding',
    long_description=open('README.md').read(),
    install_requires=[
        'networkx==2.8.4',
        'matplotlib==3.5.2',
        'stim==1.12.0',
        'pymatching==2.1.0',
        'sinter==1.11.0',
        'scipy==1.9.1',
        'stim==1.12.0',
        'numpy==1.21.6',
    ],
)


import stim

class LogicalCircuitMBQC(stim.Circuit):
    def __init__(self, circuit, loss_probabilities, potential_lost_qubits):
        super().__init__()
        for instruction in circuit:
            self.append(instruction)
        self.loss_probabilities = loss_probabilities
        self.potential_lost_qubits = potential_lost_qubits


import stim
import numpy as np
import os
import qec
import pymatching, stim
from BiasedErasure.main_code.LogicalCircuit import LogicalCircuit 
import random
import matplotlib.pyplot as plt
import progressbar
from BiasedErasure.delayed_erasure_decoders.HeraldedCircuit_FREE_LD import HeraldedCircuit_FREE_LD
from BiasedErasure.delayed_erasure_decoders.HeraldedCircuit_SWAP_LD import HeraldedCircuit_SWAP_LD
from BiasedErasure.main_code.Simulator import *
import time
import sys
from collections import defaultdict
from distutils.util import strtobool
from BiasedErasure.main_code.noise_channels import biased_erasure_noise, biased_erasure_noise_correlated, no_noise, atom_array


def create_file_name(Meta_params, bloch_point_params):
    # if Meta_params['circuit_type'] in ['Steane_Regular', 'Steane_SWAP']:  # TODO: add all the extra handles for Steane.
    #     return f"{Meta_params['architecture']}__{Meta_params['code']}__{Meta_params['circuit_type']}__{Meta_params['num_logicals']}logicals__{Meta_params['logical_basis']}__{Meta_params['bias_preserving_gates']}__{Meta_params['noise']}__{Meta_params['is_erasure_biased']}__{bloch_point_params['erasure_ratio']}__{bloch_point_params['bias_ratio']}__LD_freq_{Meta_params['LD_freq']}__SSR_{Meta_params['SSR']}__LD_method_{Meta_params['LD_method']}__num_cycles_{Meta_params['cycles']}__ordering_{Meta_params['ordering']}__decoder_{Meta_params['decoder']}"
    # elif Meta_params['circuit_type'] == 'memory' and post_selection_prob > 0:
    #     return f"{Meta_params['architecture']}__{Meta_params['code']}__{Meta_params['circuit_type']}__{Meta_params['num_logicals']}logicals__{Meta_params['logical_basis']}__{Meta_params['bias_preserving_gates']}__{Meta_params['noise']}__{Meta_params['is_erasure_biased']}__{bloch_point_params['erasure_ratio']}__{bloch_point_params['bias_ratio']}__LD_freq_{Meta_params['LD_freq']}__SSR_{Meta_params['SSR']}__LD_method_{Meta_params['LD_method']}__num_cycles_{Meta_params['cycles']}__ordering_{Meta_params['ordering']}__decoder_{Meta_params['decoder']}"
    # else:
    return f"{Meta_params['architecture']}__{Meta_params['code']}__{Meta_params['circuit_type']}__{Meta_params['num_logicals']}logicals__{Meta_params['logical_basis']}__{Meta_params['bias_preserving_gates']}__{Meta_params['noise']}__{Meta_params['is_erasure_biased']}__{bloch_point_params['erasure_ratio']}__{bloch_point_params['bias_ratio']}__LD_freq_{Meta_params['LD_freq']}__SSR_{Meta_params['SSR']}__LD_method_{Meta_params['LD_method']}__num_cycles_{Meta_params['cycles']}__ordering_{Meta_params['ordering']}__loss_decoder_{Meta_params['loss_decoder']}__decoder_{Meta_params['decoder']}"



def Bloch_sphere_point_simulate(Meta_params={}, bloch_point_params={}, output_dir='', phys_err_vec = [], distances=[5,7,9], num_shots=1000, first_comb_weight=0.5, dont_use_loss_decoder=False):
    """_summary_
    Given a set of params (meta_params for a type of a bloch sphere, and bloch_point_params for the point within), it is calling simulator
    to collect the data and save it with the file name that have all the important information.
    Args:
        Meta_params (dict, optional): Specify the bloch sphere type. Defaults to {}.
        bloch_point_params (dict, optional): Specify the R and Bias location within the sphere. Defaults to {}.
        output_dir (str, optional): Folder to save the data. Defaults to ''.
        phys_err_vec (list, optional): physical error vector. Defaults to [].
        distances (list, optional): distances. Defaults to [5,7,9].
        num_shots (int, optional): number of shots. Defaults to 1000.
    """
    file_name = create_file_name(Meta_params, bloch_point_params)
    
    noise_dict = {'biased_erasure_noise': biased_erasure_noise, 'no_noise': no_noise, 'biased_erasure_noise_correlated': biased_erasure_noise_correlated, 'atom_array': atom_array}
    noise = noise_dict[Meta_params['noise']]
    atom_array_sim = True if Meta_params['noise'] == 'atom_array' else False
    loss_detection_method_dict = {'FREE': HeraldedCircuit_FREE_LD, 'SWAP': HeraldedCircuit_SWAP_LD, 'MBQC': HeraldedCircuit_FREE_LD}
    loss_detection_method = loss_detection_method_dict[Meta_params['LD_method']]
    
    if Meta_params['code'] not in ['Rotated_XZZX', 'Rotated_Surface']: 
        Meta_params['ordering'] = None # not relevant
        
    cycles_str = Meta_params['cycles']; cycles = None if (cycles_str == 'd' or cycles_str is None) else int(cycles_str)
    
    
    simulator = Simulator(Meta_params=Meta_params, 
                        noise=noise, atom_array_sim=atom_array_sim,
                        bloch_point_params=bloch_point_params, 
                        phys_err_vec=phys_err_vec,
                        loss_detection_method=loss_detection_method, cycles = cycles,
                        output_dir=output_dir, save_filename=file_name, first_comb_weight=first_comb_weight, dont_use_loss_decoder=dont_use_loss_decoder)

    simulator.simulate(distances=distances, num_shots=num_shots)



if __name__ == '__main__':
    pass


from BiasedErasure.main_code.StabilizerCode import StabilizerCode
from BiasedErasure.main_code.xzzx_circuit import xzzx_circuit, data_ids, ancilla_ids, get_data_ancilla_indices
from BiasedErasure.main_code.xzzx_circuit_noisy import xzzx_circuit_noisy
from BiasedErasure.main_code.xzzx_cluster import xzzx_cluster, cluster_data_ids, get_neighbors_next_previous_layers, layer_offset, cluster_qubits_ids, get_num_layers
from BiasedErasure.main_code.xzzx_cluster_noisy import xzzx_cluster_noisy
import numpy as np
from BiasedErasure.main_code.noise_channels import atom_array
# from xzzx_cluster_noisy import xzzx_cluster, cluster_data_ids, get_neighbors_next_previous_layers, layer_offset, cluster_qubits_ids


class XZZX(StabilizerCode):
    def __init__(self, **kwargs):
        self.stabilizers = self.get_stabilizers()

    def get_stabilizers(self):
        # TODO: define these.
        pass
    
    def get_data_ancilla_indices(self, dx, dy, cycles, architecture):
        if architecture == 'MBQC':
            num_layers = get_num_layers(cycles)
            data_qubits, ancilla_qubits = cluster_qubits_ids(dx=dx, dy=dy, num_layers=num_layers)
            return data_qubits,  ancilla_qubits
        elif architecture == 'CBQC':
            data_qubits, ancilla_qubits = data_ids(dx,dy), ancilla_ids(dx,dy)
            return data_qubits, ancilla_qubits
        

    def get_neighbors(self, qbt, dx, dy):
        # Get neighbors of qubit qbt
        # Transform qbt to (i, j) coordinate.
        ValueError('architecture' == 'CBQC')
        alpha = 2*dx - 1
        j = qbt // alpha
        i = qbt % alpha

        neighbors = []
        if j != 2 * dy - 2:
            neighbors.append(int(((j+1)*(2*dx-1)+i)))  # up
        if i != 2 * dx - 2:
            neighbors.append(int((j*(2*dx-1)+(i+1))))  # right
        if i != 0:
            neighbors.append(int((j*(2*dx-1)+(i - 1))))  # left
        if j != 0:
            neighbors.append(int(((j-1)*(2*dx-1)+i)))  # down
        return neighbors
    
    def get_ordered_neighbors(self, qbt: int, dx: int, dy: int):
        # Get neighbors of qubit qbt in the form: {neighbor1: order1, ...}
        # where order_i specifies the time step that qbt interetacts with  neighbor_i.

        # Transform qbt to (i, j) coordinate.
        ValueError('architecture' == 'CBQC')
        alpha = 2*dx - 1
        j = qbt // alpha
        i = qbt % alpha
        neighbors = {}
        if j != 2 * dy - 2:
            neighbors[int(((j+1)*(2*dx-1)+i))] = 1  # up
        if i != 2 * dx - 2:
            neighbors[int((j*(2*dx-1)+(i+1)))] = 2  # right
        if i != 0:
            neighbors[int((j*(2*dx-1)+(i - 1)))] = 3  # left
        if j != 0:
            neighbors[int(((j-1)*(2*dx-1)+i))] = 4  # down
        return neighbors
    
    def get_order(self, direction: str):
        ordering_dict = {'up': 1, 'right': 2, 'left': 3, 'down': 4}
        return ordering_dict[direction]
        # order: up, right, left, down
        
    def get_neighbors_CZ(self, qbt, dx, dy, cycles:int = None, architecture='CBQC', gates_dicts = {}):
        # Get neighbors of qubit qbt
        # CZ neighbors from up and down (in MBQC, only in even layers and also from next and previous layers)
        
        if architecture == 'CBQC':
            # for CBQC we return a dictionary, for every qubit index (key), the order of the gate (value).
            neighbors = {1: [], 2: [], 3:[], 4:[]}
            # Transform qbt to (i, j) coordinate.
            alpha = 2*dx - 1
            j = qbt // alpha
            i = qbt % alpha
            if j != 2 * dy - 2:
                ordering_key = self.get_order('up')
                neighbors[ordering_key].append(int(((j+1)*(2*dx-1)+i)))  # up
            if j != 0:
                ordering_key = self.get_order('down')
                neighbors[ordering_key].append(int(((j-1)*(2*dx-1)+i)))  # down
        elif architecture == 'MBQC': # add neighbors from previous and next layer
            num_layers = get_num_layers(cycles)
            potential_neighbors = []
            m = int(0.5*(2*dx-1)*(2*dy-1))  # num ancilla qubits in a surface code
            n = int(np.ceil(0.5*(2*dx-1)*(2*dy-1)))  # num data qubits in a layer
            single_layer_offset = m + n
            layer_ix = qbt // single_layer_offset
            offset = layer_offset(dx=dx, dy=dy, layer_ix=layer_ix, physical=False)
            alpha = 2*dx - 1
            j = (qbt - offset) // alpha
            i = (qbt - offset) % alpha
            ValueError (qbt != j*(2*dx-1)+i+offset)
            if j != 2 * dy - 2: # north
                potential_neighbors.append((j+1)*(2*dx-1)+i+offset)
            if j != 0: # south
                potential_neighbors.append((j-1)*(2*dx-1)+i+offset)
            # other_layers_neighbors = get_neighbors_next_previous_layers(dx=dx, dy=dy, qubit_index = qbt)
            # for neighbor in other_layers_neighbors:
            #     potential_neighbors.append(neighbor)
            data_qubits, ancilla_qubits = cluster_qubits_ids(dx = dx, dy = dy, num_layers=num_layers)
            all_qubits_indices = data_qubits + ancilla_qubits
            neighbors = list(set(potential_neighbors) & set(all_qubits_indices))

        return neighbors

    def get_neighbors_CX(self, qbt, dx, dy, cycles: int = None, architecture='CBQC', gates_dicts={}):
        # Get neighbors of qubit qbt
        # CX from right and left
        if architecture == 'CBQC':
            # for CBQC we return a dictionary, for every qubit index (key), the order of the gate (value).
            neighbors = {1: [], 2: [], 3:[], 4:[]}
            # Transform qbt to (i, j) coordinate.
            alpha = 2*dx - 1
            j = qbt // alpha
            i = qbt % alpha
            if i != 2 * dx - 2:
                ordering_key = self.get_order('right')
                neighbors[ordering_key].append(int((j*(2*dx-1)+(i+1))))  # right
            if i != 0:
                ordering_key = self.get_order('left')
                neighbors[ordering_key].append(int((j*(2*dx-1)+(i - 1))))  # left
        elif architecture == 'MBQC': # add neighbors from previous and next layer
            num_layers = get_num_layers(cycles)
            potential_neighbors = []
            m = int(0.5*(2*dx-1)*(2*dy-1))  # num ancilla qubits in a surface code
            n = int(np.ceil(0.5*(2*dx-1)*(2*dy-1)))  # num data qubits in a layer
            single_layer_offset = m + n
            layer_ix = qbt // single_layer_offset
            offset = layer_offset(dx=dx, dy=dy, layer_ix=layer_ix, physical=False)
            alpha = 2*dx - 1
            j = (qbt - offset) // alpha
            i = (qbt - offset) % alpha
            ValueError (qbt != j*(2*dx-1)+i+offset)
            if i != 2 * dx - 2: # east
                potential_neighbors.append(j*(2*dx-1)+(i+1)+offset)
            if i != 0: # west
                potential_neighbors.append(j*(2*dx-1)+(i - 1)+offset)            
            # new addition - other layers:
            other_layers_neighbors = get_neighbors_next_previous_layers(dx=dx, dy=dy, qubit_index = qbt, num_layers=num_layers)
            for neighbor in other_layers_neighbors:
                potential_neighbors.append(neighbor)
            data_qubits, ancilla_qubits = cluster_qubits_ids(dx = dx, dy = dy, num_layers=num_layers)
            all_qubits_indices = data_qubits + ancilla_qubits
            # all_qubits_indices = cluster_qubits_ids(dx = dx, dy = dy, num_layers=num_layers)
            neighbors = list(set(potential_neighbors) & set(all_qubits_indices))
        
        return neighbors
    
    
    def transfer_qubit_number_to_ij_index(self, qbt, dx, dy):
        alpha = 2*dx - 1
        j = qbt // alpha
        i = qbt % alpha
        return [i,j]
    
    def get_down_neighbor(self, qbt, dx, dy):
        # Transform qbt to (i, j) coordinate.
        alpha = 2*dx - 1
        j = qbt // alpha
        i = qbt % alpha
        if j != 0:
            return int(((j-1)*(2*dx-1)+i))  # down
        else:
            return None
        
    def get_up_neighbor(self, qbt, dx, dy):
        # Transform qbt to (i, j) coordinate.
        alpha = 2*dx - 1
        j = qbt // alpha
        i = qbt % alpha
        if j != 2 * dy - 2:
            return int(((j+1)*(2*dx-1)+i))  # up
        else:
            return None
        

    def get_circuit(self, cycles, dx: int, dy: int, basis='Z', architecture='CBQC', atom_array_sim = False, **kwargs):
        if architecture == 'MBQC':
            if atom_array_sim:
                return xzzx_cluster_noisy(cycles=cycles, dx=dx, dy=dy, basis=basis, **kwargs)
            else:
                return xzzx_cluster(cycles=cycles, dx=dx, dy=dy, basis=basis, **kwargs)
            
        elif architecture == 'CBQC':
            if atom_array_sim:
                return xzzx_circuit_noisy(cycles=cycles, dx=dx, dy=dy, basis=basis, **kwargs)
            else:
                return xzzx_circuit(cycles=cycles, dx=dx, dy=dy, basis=basis, **kwargs)

        else:
            assert True == False
    
    def get_all_data_ids(self, architecture, dx, dy, cycles:int):
        if architecture == "MBQC":
            return cluster_data_ids(dx,dy,cycles)
        elif architecture == "CBQC":
            return data_ids(dx,dy)


    def get_all_ancilla_ids(self, architecture, dx: int, dy: int, cycles: int = -1):
        if architecture == "MBQC":
            # TODO: implement this.
            return None
        elif architecture == "CBQC":
            return ancilla_ids(dx, dy)


    def get_data_neighbors_of_ancilla_cbqc(self, dx: int, dy: int, **kwargs):
        ancillas = ancilla_ids(dx=dx, dy=dy) # GB: I changed it, to fix the "architecture is missing" bug
        # ancillas = self.get_all_ancilla_ids(dx=dx, dy=dy, architecture=architecture)
        data_neighbors = {ancilla: self.get_ordered_neighbors(ancilla, dx, dy) for ancilla in ancillas}
        return data_neighbors


    
    
    def get_all_qubit_indices(self, dx, dy, cycles, architecture):
        if architecture == 'MBQC':
            num_layers = get_num_layers(cycles)
            data_qubits, ancilla_qubits = cluster_qubits_ids(dx, dy, num_layers=num_layers)
            return data_qubits +  ancilla_qubits
        elif architecture == 'CBQC':
            data_qubits, ancilla_qubits = get_data_ancilla_indices(dx,dy)
            return data_qubits +  ancilla_qubits

    
    def get_all_qubit_indices_layer(self, dx, dy, layer_ix, architecture, single_list = False):
        if architecture == 'MBQC':
            offset = layer_offset(dx, dy, layer_ix)
            ancilla_qubits = [j*(2*dx-1)+i + offset for j in range(2*dy-1) for i in range(2*dx-1) if ((i+j)%2 == 1) and ((i%2 ==0 and layer_ix%2==0) or (i%2 ==1 and layer_ix%2==1))]
            data_qubits = [j*(2*dx-1)+i + offset for j in range(2*dy-1) for i in range(2*dx-1) if ((i+j)%2 == 0)]
            if single_list:
                return data_qubits + ancilla_qubits
            else:
                return [data_qubits, ancilla_qubits]
        elif architecture == 'CBQC':
            print(f"Need to build this feature")
            return None
        
        
        
    def get_gates_dicts(self, architecture, dx, dy):
        if architecture == "MBQC":
            print(f"Need to build this feature")
            return None
        elif architecture == "CBQC":
            print(f"Need to build this feature")
            return None

    def get_initialization_basis(self, qbt, dx, dy, basis='X'):
        alpha = 2*dx - 1
        j = qbt // alpha
        i = qbt % alpha
        (prep1, prep2) = ('RZ', 'RX') if basis == 'X' else ('RX', 'RZ')
        assert (i+j) % 2 == 0
        if i%2 == 0:
            return prep1
        else:
            return prep2
    
    def find_SWAP_undetected_qubits(self, dx, dy, SWAP_round_ix):
        # only relevant for CBQC
        data_qubits = data_ids(dx,dy)
        unheralded_data_qubits = []
        for qbt in data_qubits:
            alpha = 2*dx - 1
            j = qbt // alpha
            i = qbt % alpha
            if (SWAP_round_ix%2 == 0 and j == 0) or (SWAP_round_ix%2 == 1 and j == 2*dy-2):
                unheralded_data_qubits.append(qbt)
        return unheralded_data_qubits
                
        
        